"""Testing DrugAffinityDataset."""
import os
import unittest

import numpy as np
from torch.utils.data import DataLoader

from pytoda.datasets import DrugAffinityDataset
from pytoda.tests.utils import TestFileContent

COLUMN_NAMES = [',ligand_name,sequence_id,label', ',drug,protein,class']
DRUG_AFFINITY_CONTENT = os.linesep.join(
    [
        '0,CHEMBL14688,name=20S proteasome chymotrypsin-like-organism=Homo sapiens,1',  # noqa
        '1,CHEMBL14688,name=21S proteasome chymotrypsin-like-organism=Homo sapiens,0',  # noqa
        '2,CHEMBL17564,name=20S proteasome chymotrypsin-like-organism=Homo sapiens,0',  # noqa
        '3,CHEMBL17564,name=21S proteasome chymotrypsin-like-organism=Homo sapiens,1',  # noqa
    ]
)
SMILES_CONTENT = os.linesep.join(
    ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
)
PROTEIN_SEQUENCE_CONTENT = os.linesep.join(
    [
        'ABC\tname=20S proteasome chymotrypsin-like-organism=Homo sapiens',  # noqa
        'DEFG\tname=21S proteasome chymotrypsin-like-organism=Homo sapiens',  # noqa
        'CDEF\tname=22S proteasome chymotrypsin-like-organism=Homo sapiens',  # noqa
        'XZSDASDFF\tname=23S proteasome chymotrypsin-like-organism=Homo sapiens',  # noqa
    ]
)


class TestDrugAffinityDatasetEagerBackend(unittest.TestCase):
    """Testing DrugAffinityDataset with eager backend."""

    def setUp(self):
        self.backend = 'eager'
        print(f'backend is {self.backend}')
        self.smiles_content = SMILES_CONTENT
        self.protein_sequence_content = PROTEIN_SEQUENCE_CONTENT

        for column_names in COLUMN_NAMES:
            self.drug_affinity_content = os.linesep.join(
                [column_names, DRUG_AFFINITY_CONTENT]
            )

            with TestFileContent(self.drug_affinity_content) as drug_affinity_file:
                with TestFileContent(self.smiles_content) as smiles_file:
                    with TestFileContent(
                        self.protein_sequence_content
                    ) as protein_sequence_file:
                        self.drug_affinity_dataset = DrugAffinityDataset(
                            drug_affinity_file.filename,
                            smiles_file.filename,
                            protein_sequence_file.filename,
                            backend=self.backend,
                            column_names=column_names.split(',')[1:],
                        )

    def test___len__(self) -> None:
        """Test __len__."""
        self.assertEqual(len(self.drug_affinity_dataset), 4)

    def test___getitem__(self) -> None:
        """Test __getitem__."""
        smiles_padding_index = (
            self.drug_affinity_dataset.smiles_dataset.smiles_language.padding_index
        )
        smiles_c_index = (
            self.drug_affinity_dataset.smiles_dataset.smiles_language.token_to_index[
                'C'
            ]
        )
        smiles_o_index = (
            self.drug_affinity_dataset.smiles_dataset.smiles_language.token_to_index[
                'O'
            ]
        )
        protein_sequence_padding_index = (
            self.drug_affinity_dataset.protein_sequence_dataset.protein_language.padding_index
        )
        protein_sequence_a_index = self.drug_affinity_dataset.protein_sequence_dataset.protein_language.token_to_index[
            'A'
        ]
        protein_sequence_b_index = self.drug_affinity_dataset.protein_sequence_dataset.protein_language.token_to_index[
            'B'
        ]
        protein_sequence_c_index = self.drug_affinity_dataset.protein_sequence_dataset.protein_language.token_to_index[
            'C'
        ]
        (
            smiles_indexes_tensor,
            protein_sequence_indexes_tensor,
            label_tensor,
        ) = self.drug_affinity_dataset[0]
        np.testing.assert_almost_equal(
            smiles_indexes_tensor.numpy(),
            np.array(
                [
                    smiles_padding_index,
                    smiles_padding_index,
                    smiles_c_index,
                    smiles_o_index,
                ]
            ),
        )
        np.testing.assert_almost_equal(
            protein_sequence_indexes_tensor.numpy(),
            np.array(
                6 * [protein_sequence_padding_index]
                + [
                    protein_sequence_a_index,
                    protein_sequence_b_index,
                    protein_sequence_c_index,
                ]
            ),
        )
        np.testing.assert_almost_equal(label_tensor.numpy(), np.array([1], dtype=int))

    def test_data_loader(self) -> None:
        """Test data_loader."""
        data_loader = DataLoader(self.drug_affinity_dataset, batch_size=2, shuffle=True)
        for (
            batch_index,
            (smiles_indexes_batch, protein_sequence_indexes_batch, label_batch),
        ) in enumerate(data_loader):
            self.assertEqual(smiles_indexes_batch.size(), (2, 4))
            self.assertEqual(protein_sequence_indexes_batch.size(), (2, 9))
            self.assertEqual(label_batch.size(), (2, 1))
            if batch_index > 4:
                break


class TestDrugAffinityDatasetLazyBackend(TestDrugAffinityDatasetEagerBackend):
    """Testing DrugAffinityDataset with lazy backend."""

    def setUp(self):
        self.backend = 'lazy'
        print(f'backend is {self.backend}')
        self.smiles_content = SMILES_CONTENT
        self.protein_sequence_content = PROTEIN_SEQUENCE_CONTENT

        for column_names in COLUMN_NAMES:
            self.drug_affinity_content = os.linesep.join(
                [column_names, DRUG_AFFINITY_CONTENT]
            )

            with TestFileContent(self.drug_affinity_content) as drug_affinity_file:
                with TestFileContent(self.smiles_content) as smiles_file:
                    with TestFileContent(
                        self.protein_sequence_content
                    ) as protein_sequence_file:
                        self.drug_affinity_dataset = DrugAffinityDataset(
                            drug_affinity_file.filename,
                            smiles_file.filename,
                            protein_sequence_file.filename,
                            backend=self.backend,
                            column_names=column_names.split(',')[1:],
                        )


if __name__ == '__main__':
    unittest.main()



"""Implementation of _FastaEagerDataset."""
from upfp import parse_fasta

from ..types import Hashable
from .base_dataset import KeyDataset


class _FastaEagerDataset(KeyDataset):
    """
    .fasta dataset using eager loading.

    Suggested when handling datasets that can fit in the device memory.

    Only supports uniprot fasta headers
    """

    def __init__(
        self, fasta_filepath: str, gzipped=True, name: str = 'Sequence'
    ) -> None:
        """
        Initialize a .fasta dataset.

        Args:
            fasta_filepath (str): path to .fasta file.
            gzipped (bool): Whether or not fasta file is zipped (.fasta.gz).
            name (str): type of dataset, used to index columns.
        """
        KeyDataset.__init__(self)
        self.fasta_filepath = fasta_filepath
        self.name = name

        self.key_to_index_mapping = {}
        self.ordered_keys = []
        self.samples = []
        try:
            for index, item in enumerate(parse_fasta(fasta_filepath, gzipped=gzipped)):
                key = item['accession_number']  # uniprot unique identifier
                self.key_to_index_mapping[key] = index
                self.ordered_keys.append(key)
                self.samples.append(item['sequence'])
        except KeyError:
            raise KeyError('Badly formatted .fasta file, no sequence found.')

    def __len__(self) -> int:
        """Total number of samples."""
        return len(self.samples)

    def __getitem__(self, index: int) -> str:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            str: a sequence in the FASTA file.
        """
        return self.samples[index]

    def get_key(self, index: int) -> Hashable:
        """Get key from integer index."""
        return self.ordered_keys[index]

    def get_index(self, key: Hashable) -> int:
        """Get index for first datum mapping to the given key."""
        return self.key_to_index_mapping[key]

    @property
    def has_duplicate_keys(self):
        return len(self.ordered_keys) != len(self.key_to_index_mapping)

    def keys(self):
        return iter(self.ordered_keys)



"""Type definitions."""
import inspect
from collections import OrderedDict  # noqa
from typing import (  # noqa
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from numpy import ndarray  # np.typing in development
from pandas import DataFrame, Series
from torch import Tensor

Tokens = List[str]
Indexes = List[int]
Tokenizer = Callable[[str], Tokens]
Files = Sequence[str]  # often passed on as Tuple via *args
FeatureList = List[str]
GeneList = FeatureList
CsvSourceData = Union[ndarray, Series, DataFrame]
CallableOnSource = Union[
    Callable[[ndarray], ndarray],
    Callable[[Series], Series],
    Callable[[DataFrame], DataFrame],
]
TransformList = List[Callable[[Any], Any]]
DrugSensitivityData = Tuple[Tensor, Tensor, Tensor]
DrugSensitivityDoseData = Tuple[Tensor, Tensor, Tensor, Tensor]
DrugAffinityData = Tuple[Tensor, Tensor, Tensor]
AnnotatedData = Tuple[Any, Tensor]


def delegate_kwargs(to=None, keep=False):
    """
    Decorator: replace `**kwargs` in signature with params from `to`.

    Source: https://www.fast.ai/2019/08/06/delegation/
    """

    def _f(f):
        if to is None:
            to_f, from_f = f.__base__.__init__, f.__init__
        else:
            to_f, from_f = to, f
        sig = inspect.signature(from_f)
        sigd = dict(sig.parameters)
        k = sigd.pop('kwargs')
        s2 = {
            k: v
            for k, v in inspect.signature(to_f).parameters.items()
            if v.default != inspect.Parameter.empty and k not in sigd
        }
        sigd.update(s2)
        if keep:
            sigd['kwargs'] = k
        from_f.__signature__ = sig.replace(parameters=sigd.values())
        return f

    return _f



"""Testing GeneExpressionDataset with eager backend."""
import os
import unittest

import numpy as np
import pandas as pd
from torch.utils.data import DataLoader

from pytoda.datasets import GeneExpressionDataset
from pytoda.tests.utils import TestFileContent

CONTENT = os.linesep.join(
    [
        'genes,A,C,B,D',
        'sample_3,9.45,4.984,7.016,8.336',
        'sample_2,7.188,0.695,10.34,6.047',
        'sample_1,9.25,6.133,5.047,5.6',
    ]
)
MORE_CONTENT = os.linesep.join(
    [
        'genes,B,C,D,E,F',
        'sample_10,4.918,0.0794,1.605,3.463,10.18',
        'sample_11,3.043,8.56,1.961,0.6226,5.027',
        'sample_12,4.76,1.124,6.06,0.3743,11.05',
        'sample_13,0.626,5.164,4.277,4.414,2.7',
    ]
)


class TestGeneExpressionDatasetEagerBackend(unittest.TestCase):
    """Testing GeneExpressionDataset with eager backend."""

    def setUp(self):
        self.backend = 'eager'
        print(f'backend is {self.backend}')
        self.content = CONTENT
        self.other_content = MORE_CONTENT

    def test___len__(self) -> None:
        """Test __len__."""

        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:
                gene_expression_dataset = GeneExpressionDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                    index_col=0,
                )
                self.assertEqual(len(gene_expression_dataset), 7)

    def test___getitem__(self) -> None:
        """Test __getitem__."""
        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:
                df = pd.concat(
                    [
                        pd.read_csv(a_test_file.filename, index_col=0),
                        pd.read_csv(another_test_file.filename, index_col=0),
                    ],
                    sort=False,
                )
                gene_expression_dataset = GeneExpressionDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                    index_col=0,
                )
                gene_list = gene_expression_dataset.gene_list
                mean = df.mean()[gene_list].values
                std = df.std(ddof=0)[gene_list].values
                for i, (key, row) in enumerate(df[gene_list].iterrows()):
                    np.testing.assert_almost_equal(
                        gene_expression_dataset[i].numpy(), (row.values - mean) / std, 5
                    )
                np.testing.assert_almost_equal(gene_expression_dataset.mean, mean, 5)
                np.testing.assert_almost_equal(gene_expression_dataset.std, std, 5)

                gene_expression_dataset = GeneExpressionDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                    index_col=0,
                    standardize=False,
                    min_max=True,
                )
                minimum = df.min()[gene_list].values
                maximum = df.max()[gene_list].values
                diff = maximum - minimum
                for i, (key, row) in enumerate(df[gene_list].iterrows()):
                    np.testing.assert_almost_equal(
                        gene_expression_dataset[i].numpy(),
                        (row.values - minimum) / diff,
                        5,
                    )
                np.testing.assert_almost_equal(gene_expression_dataset.min, minimum, 5)
                np.testing.assert_almost_equal(gene_expression_dataset.max, maximum, 5)

    def test_processing_parameters_standardize_reindex(self) -> None:
        with TestFileContent(self.content) as a_test_file, TestFileContent(
            self.other_content
        ) as another_test_file:
            # feature not in data is filled with zeros
            feature_list = ['E', 'C', 'D', 'B', 'all_missing']
            standard_dataset = GeneExpressionDataset(
                a_test_file.filename,
                another_test_file.filename,
                gene_list=feature_list,
                backend=self.backend,
                index_col=0,
            )
            self.assertEqual(standard_dataset[0][-1], 0)

            gene_list = standard_dataset.gene_list
            df = pd.concat(
                [
                    pd.read_csv(a_test_file.filename, index_col=0),
                    pd.read_csv(another_test_file.filename, index_col=0),
                ],
                sort=False,
            ).reindex(
                columns=gene_list
            )  # , fill_value=0.0)

            # scalar scaling (single max and min)
            flat = df.values.flatten()
            # allow nan
            mean_float = np.nanmean(flat)
            std_float = np.nanstd(flat)
            for mean, std in [
                # scalar
                [mean_float, std_float],
                # list length 1
                [[mean_float], [std_float]],
            ]:
                processing_parameters = {
                    'mean': mean,
                    'std': std,
                }
                standard_ds = GeneExpressionDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    gene_list=feature_list,
                    backend=self.backend,
                    index_col=0,
                    standardize=True,
                    min_max=False,
                    processing_parameters=processing_parameters,
                    impute=None,
                )

                # collect flat values
                ds_1d = np.concatenate([item.numpy() for item in standard_ds])
                # allowing/ignoring nan
                np.testing.assert_almost_equal(np.nanmean(ds_1d), 0)
                np.testing.assert_almost_equal(np.nanstd(ds_1d), 1)

            mean_array = df.mean().values
            std_array = df.std(ddof=0).values
            # NOTE: numpy and pytoda use ddof of 0, whereas pandas default is 1
            for mean, std in [
                # list
                [mean_array.tolist(), std_array.tolist()],
                # ndarray
                [mean_array, std_array],
            ]:
                processing_parameters = {
                    'mean': mean,
                    'std': std,
                }
                standard_ds = GeneExpressionDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    gene_list=feature_list,
                    backend=self.backend,
                    index_col=0,
                    standardize=True,
                    min_max=False,
                    processing_parameters=processing_parameters,
                    impute=None,
                )

                # collect transformed values
                ds_2d = np.stack([item.numpy() for item in standard_ds])
                # contains nan
                ds_means = np.nanmean(ds_2d, axis=0)
                ds_stds = np.nanstd(ds_2d, axis=0)

                # debug std
                # standard_ds.std
                # std_array        # == df.std(ddof=0).values
                # df.std().values  # == df.std(ddof=1).values

                for index, feature in enumerate(standard_ds.gene_list):
                    if feature == 'all_missing':
                        # no features at all so transformed stat is nan
                        self.assertTrue(np.isnan(ds_means[index]))
                        self.assertTrue(np.isnan(ds_stds[index]))
                        # original stats are also nan
                        np.testing.assert_almost_equal(
                            standard_ds.mean[index], df[feature].mean()
                        )
                        np.testing.assert_almost_equal(
                            standard_ds.std[index], df[feature].std(ddof=0)
                        )
                        continue

                    # note some NaN values 'E' still have statistics
                    # TODO our reduce statistics has to cope with some NaN!
                    # until then 'E' fails

                    # check transformed std / mean are 1 and 0 per feature
                    np.testing.assert_almost_equal(ds_means[index], 0)
                    np.testing.assert_almost_equal(ds_stds[index], 1)
                    # external statisic matches internally used statistics
                    np.testing.assert_almost_equal(
                        standard_ds.mean[index], df[feature].mean()
                    )
                    np.testing.assert_almost_equal(
                        standard_ds.std[index], df[feature].std(ddof=0)
                    )
                    # order of reduced means matches order in

    def test_processing_parameters_minmax(self) -> None:
        with TestFileContent(self.content) as a_test_file, TestFileContent(
            self.other_content
        ) as another_test_file:
            minmax_dataset = GeneExpressionDataset(
                a_test_file.filename,
                another_test_file.filename,
                backend=self.backend,
                index_col=0,
                standardize=False,
                min_max=True,
            )
            gene_list = minmax_dataset.gene_list
            df = pd.concat(
                [
                    pd.read_csv(a_test_file.filename, index_col=0),
                    pd.read_csv(another_test_file.filename, index_col=0),
                ],
                sort=False,
            )[gene_list]

            # with min max scaling we can check for values 0 and 1
            maximum_array = df.max().values
            minimum_array = df.min().values

            # scalar scaling (single max and min)
            max_n, max_p = map(int, np.unravel_index(np.argmax(df.values), df.shape))
            min_n, min_p = map(int, np.unravel_index(np.argmin(df.values), df.shape))
            for maximum, minimum in [
                # scalar
                [np.max(maximum_array), np.min(minimum_array)],
                # list length 1
                [[np.max(maximum_array)], [np.min(minimum_array)]],
            ]:
                processing_parameters = {
                    'max': maximum,
                    'min': minimum,
                }
                minmax_ds = GeneExpressionDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                    index_col=0,
                    standardize=False,
                    min_max=True,
                    processing_parameters=processing_parameters,
                )

                self.assertEqual(minmax_ds[max_n][max_p], 1)
                self.assertEqual(minmax_ds[min_n][min_p], 0)

            # array scaling (feature wise max and min)
            max_indeces = map(int, np.argmax(df.values, axis=0))
            min_indeces = map(int, np.argmin(df.values, axis=0))

            for maximum, minimum in [
                # list
                [maximum_array.tolist(), minimum_array.tolist()],
                # ndarray
                [maximum_array, minimum_array],
            ]:
                processing_parameters = {
                    'max': maximum,
                    'min': minimum,
                }
                minmax_ds = GeneExpressionDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                    index_col=0,
                    standardize=False,
                    min_max=True,
                    processing_parameters=processing_parameters,
                )
                # check max_index / min_index are 1 and 0 per feature
                for feature_index, sample_index in enumerate(max_indeces):
                    self.assertEqual(minmax_ds[sample_index][feature_index], 1)
                for feature_index, sample_index in enumerate(min_indeces):
                    self.assertEqual(minmax_ds[sample_index][feature_index], 0)

    def test_data_loader(self) -> None:
        """Test data_loader."""
        gene_subset_list = ['B', 'D', 'F']
        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:
                gene_expression_dataset = GeneExpressionDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    gene_list=gene_subset_list,
                    backend=self.backend,
                    index_col=0,
                )
                data_loader = DataLoader(
                    gene_expression_dataset, batch_size=2, shuffle=True
                )
                for batch_index, batch in enumerate(data_loader):
                    self.assertEqual(
                        batch.shape,
                        (
                            1 if batch_index == 3 else 2,
                            gene_expression_dataset.number_of_features,
                        ),
                    )
                    if batch_index > 2:
                        break

    def _test_indexed(self, ds, keys, index):
        key = keys[index]
        positive_index = index % len(ds)
        # get_key (support for negative index?)
        self.assertEqual(key, ds.get_key(positive_index))
        self.assertEqual(key, ds.get_key(index))
        # get_index
        self.assertEqual(positive_index, ds.get_index(key))
        # get_item_from_key
        self.assertTrue(all(ds[index] == ds.get_item_from_key(key)))
        # keys
        self.assertSequenceEqual(keys, list(ds.keys()))
        # duplicate keys
        self.assertFalse(ds.has_duplicate_keys)

    def test_all_base_for_indexed_methods(self):

        with TestFileContent(self.content) as a_test_file:
            with TestFileContent(self.other_content) as another_test_file:
                gene_expression_dataset = GeneExpressionDataset(
                    a_test_file.filename,
                    another_test_file.filename,
                    backend=self.backend,
                    index_col=0,
                )
                gene_expression_ds_0 = GeneExpressionDataset(
                    a_test_file.filename, backend=self.backend, index_col=0
                )
                gene_expression_ds_1 = GeneExpressionDataset(
                    another_test_file.filename, backend=self.backend, index_col=0
                )
        all_keys = [
            row.split(',')[0]
            for row in self.content.split(os.linesep)[1:]
            + self.other_content.split(os.linesep)[1:]
        ]

        for ds, keys in [
            (gene_expression_dataset, all_keys),
            (gene_expression_ds_0, all_keys[: len(gene_expression_ds_0)]),
            (gene_expression_ds_1, all_keys[len(gene_expression_ds_0) :]),
            (gene_expression_ds_0 + gene_expression_ds_1, all_keys),
        ]:
            index = -1
            self._test_indexed(ds, keys, index)

        # duplicate
        duplicate_ds = gene_expression_ds_0 + gene_expression_ds_0
        self.assertTrue(duplicate_ds.has_duplicate_keys)

        # GeneExpressionDataset does not test and raise
        with TestFileContent(self.content) as a_test_file:
            gene_expression_dataset = GeneExpressionDataset(
                a_test_file.filename,
                a_test_file.filename,
                backend=self.backend,
                index_col=0,
            )
            self.assertTrue(gene_expression_dataset.has_duplicate_keys)


class TestGeneExpressionDatasetLazyBackend(
    TestGeneExpressionDatasetEagerBackend
):  # noqa
    """Testing GeneExpressionDataset with lazy backend."""

    def setUp(self):
        self.backend = 'lazy'
        print(f'backend is {self.backend}')
        self.content = CONTENT
        self.other_content = MORE_CONTENT


if __name__ == '__main__':
    unittest.main()



import logging
import urllib
import urllib.request as urllib_request
from itertools import filterfalse
from typing import Iterable, List, Tuple, Union
from urllib.error import HTTPError

import pubchempy as pcp
from pubchempy import BadRequestError, PubChemHTTPError
from unidecode import unidecode

from pytoda.smiles.transforms import Canonicalization

logger = logging.getLogger(__name__)

ZINC_DRUG_SEARCH_ROOT = 'http://zinc.docking.org/substances/search/?q='
ZINC_ID_SEARCH_ROOT = 'http://zinc.docking.org/substances/'

PUBCHEM_START = 'https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound'
PUBCHEM_MID = 'property'
PUBCHEM_END = 'TXT'


def get_smiles_from_zinc(drug: Union[str, int]) -> str:
    """
    Uses the ZINC databases to retrieve the SMILES of a ZINC ID (int) or a drug
    name (str).

    Args:
        drug (Union[str, int]): a string with a drug name or an int of a ZINC
            ID.
    Returns:
        smiles (str): The SMILES string of the drug name or ZINC ID.
    """

    if type(drug) != str and type(drug) != int:
        raise TypeError(
            f'Please insert drug of type {{str, int}}, given was {type(drug)}'
            f'({drug}).'
        )

    if type(drug) == str:

        # Parse name, then retrieve ZINC ID from it
        stripped_drug = unidecode(drug).strip().replace(' ', '%20')
        zinc_ids = []
        try:
            drug_url = urllib_request.pathname2url(stripped_drug)
            path = '{}{}'.format(ZINC_DRUG_SEARCH_ROOT, drug_url)
            response = urllib.request.urlopen(path)

            for line in response:
                line = line.decode(encoding='UTF-8').strip()
                if 'href="/substances/ZINC' in line:
                    zinc_ids.append(line.split('/')[-2])
            zinc_id = zinc_ids[0]

        except HTTPError:
            logger.warning(f'Did not find any result for drug: {drug}')
            return ''

    elif type(drug) == int:
        zinc_id = str(drug)

    zinc_id_url = ZINC_ID_SEARCH_ROOT + zinc_id
    id_response = urllib_request.urlopen(zinc_id_url)

    for id_line in id_response:
        id_line = id_line.decode(encoding='UTF-8').strip()
        if 'id="substance-smiles-field" readonly value=' in id_line:
            smiles = id_line.split('"')[-2]

    return smiles


def get_smiles_from_pubchem(
    drug: Union[str, int],
    query_type: str = 'name',
    use_isomeric: bool = True,
    kekulize: bool = False,
    sanitize: bool = True,
) -> str:
    """
    Uses the PubChem database to retrieve the SMILES of a drug name given as string
    (default) or a PubChem ID.

    Args:
        drug (str): string with a drug name (or a PubChem ID as a string).
        query_type (str): Either 'name' or 'cid'.
            Identifies whether the argument provided as drug is a name (e.g 'Tacrine') or
            a PubChem ID (1935). Defaults to name.
        use_isomeric (bool, optional) - If available, returns the isomeric
            SMILES, not the canonical one.
        kekulize (bool, optional): whether kekulization is used. PubChem uses
            kekulization per default, so setting this to 'True' will not
            perform any operation on the retrieved SMILES.
            NOTE: Setting it to 'False' will convert aromatic atoms to lower-
            case characters and *induces a RDKit dependency*
        sanitize (bool, optional): Sanitize SMILE
    Returns:
        smiles (str): The SMILES string of the drug name.
    """

    if not kekulize and not sanitize:
        raise ValueError(
            'If Kekulize is False, molecule has to be sanitize '
            '(sanitize cannot be False).'
        )

    if type(drug) != str and type(drug) != int:
        raise TypeError(
            f'Please insert drug of type str or int, given was {type(drug)}({drug}).'
        )
    if not kekulize:
        from rdkit import Chem

    options = ['CanonicalSMILES']
    if use_isomeric:
        options = ['IsomericSMILES'] + options

    # Parse name
    if isinstance(drug, str):
        drug = unidecode(drug).strip().replace(' ', '%20')

    # Search ZINC for compound name
    for option in options:
        try:
            path = '{}/{}/{}/{}/{}/{}'.format(
                PUBCHEM_START, query_type, drug, PUBCHEM_MID, option, PUBCHEM_END
            )
            smiles = (
                urllib_request.urlopen(path).read().decode('UTF-8').replace('\n', '')
            )
            if not kekulize:
                smiles = Chem.MolToSmiles(Chem.MolFromSmiles(smiles, sanitize=sanitize))
            return smiles
        except HTTPError:
            if option == 'CanonicalSMILES':
                logger.warning(f'Did not find any result for drug: {drug}')
                return ''
            continue


def remove_pubchem_smiles(smiles_list: Iterable[str]) -> List:
    """
    Function for removing PubChem molecules from an iterable of smiles.
    Args:
        smiles_list (Iterable[str]): many SMILES strings.
    Returns:
        List[str]:  Filtered list of SMILES, all SMILES pointing to PubChem
            molecules are removed.
    """

    if not isinstance(smiles_list, Iterable):
        raise TypeError(f'Please pass Iterable, not {type(smiles_list)}')

    canonicalizer = Canonicalization(sanitize=False)
    filtered = filterfalse(is_pubchem, smiles_list)
    # Canonicalize molecules and filter again (sanity check)
    filtered_canonical = filterfalse(lambda x: is_pubchem(canonicalizer(x)), filtered)
    return list(filtered_canonical)


def query_pubchem(smiles: str) -> Tuple[bool, int]:
    """
    Queries pubchem for a given SMILES.

    Args:
        smiles (str): A SMILES string.

    Returns:
        Tuple[bool, int]:
            bool: Whether or not SMILES is known to PubChem.

            int: PubChem ID of matched SMILES, -1 if SMILES was not found.
                Instead, -2 means an error in the PubChem query.
    """

    if not isinstance(smiles, str):
        raise TypeError(f'Please pass str, not {type(smiles)}')
    try:
        result = pcp.get_compounds(smiles, 'smiles')[0]
    except BadRequestError:
        logger.warning(f'Skipping SMILES. BadRequestError with: {smiles}')
        return (False, -2)
    except HTTPError:
        logger.warning(f'Skipping SMILES. HTTPError with: {smiles}')
        return (False, -2)
    except TimeoutError:
        logger.warning(f'Skipping SMILES. TimeoutError with: {smiles}')
        return (False, -2)
    except ConnectionResetError:
        logger.warning(f'Skipping SMILES. ConnectionResetError with: {smiles}')
        return (False, -2)
    except PubChemHTTPError:
        logger.warning(f'Skipping SMILES, server busy. with: {smiles}')
        return (False, -2)

    return (False, -1) if result.cid is None else (True, result.cid)


def is_pubchem(smiles: str) -> bool:
    """Whether a given SMILES in PubChem.
    Args:
        smiles (str): A SMILES string.
    Returns:
        bool: Whether or not SMILES is known to PubChem.
    """
    return query_pubchem(smiles)[0]



"""Testing data splitter."""
import os
import tempfile
import unittest
from typing import Tuple

import pandas as pd

from pytoda.data_splitter import csv_data_splitter
from pytoda.tests.utils import TestFileContent
from pytoda.types import Files


class TestDataSplitter(unittest.TestCase):
    """Testing csv data splitting."""

    def _read_dfs(self, filepaths: Files) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Read data frames from a list of files.

        Args:
            filepaths (Files): a list of files.

        Returns:
            Tuple[pd.DataFrame, ...]: a tuple of data frames.
        """
        return tuple(pd.read_csv(filepath, index_col=0) for filepath in filepaths)

    def test_data_splitter(self) -> None:
        """Test csv_data_splitter."""

        a_content = os.linesep.join(
            [
                'genes,A,B,C,D',
                'sample_3,9.45,4.984,7.016,8.336',
                'sample_2,7.188,0.695,10.34,6.047',
                'sample_1,9.25,6.133,5.047,5.6',
            ]
        )
        another_content = os.linesep.join(
            [
                'genes,B,C,D,E,F',
                'sample_10,4.918,0.0794,1.605,3.463,10.18',
                'sample_11,3.043,8.56,1.961,0.6226,5.027',
                'sample_12,4.76,1.124,6.06,0.3743,11.05',
                'sample_13,0.626,5.164,4.277,4.414,2.7',
            ]
        )

        # first row for random splits, second for file split
        ground_truth = [
            [(6, 6), (1, 6)],
            [(5, 6), (2, 6)],
            [(3, 6), (4, 6)],
            [(3, 4), (4, 5)],
            [(3, 4), (4, 5)],
            [(3, 4), (4, 5)],
        ]
        index = 0
        with tempfile.TemporaryDirectory() as directory:
            for mode in ['random', 'file']:
                for test_fraction in [0.1, 0.2, 0.5]:
                    with TestFileContent(a_content) as a_test_file:
                        with TestFileContent(another_content) as another_test_file:

                            train_filepath, test_filepath = csv_data_splitter(
                                [a_test_file.filename, another_test_file.filename],
                                directory,
                                'general',
                                mode=mode,
                                test_fraction=test_fraction,
                            )
                            train_df, test_df = self._read_dfs(
                                [train_filepath, test_filepath]
                            )
                            self.assertEqual(
                                ground_truth[index], [train_df.shape, test_df.shape]
                            )
                            index += 1


if __name__ == '__main__':
    unittest.main()



"""Utilities for the tests."""
import os
import tempfile
import warnings


class TestFileContent:
    """
    Create a temporary file with a given content.

    Inspired by: https://stackoverflow.com/a/54053967/10032558.
    """

    __test__ = False  # avoid PytestCollectionWarning

    def __init__(self, content: str, **kwargs) -> None:
        """
        Initialize the file with a content.

        Args:
            content (str): content of the file.
            **kwargs (dict): Additional keyword arguments for NamedTemporaryFile.
                NOTE: This can e.g. be suffix='.csv' if the temporary filename should
                adhere to a specific suffix.
        """
        self.file = tempfile.NamedTemporaryFile(mode='w', delete=False, **kwargs)
        with self.file as fp:
            fp.write(content)

    @property
    def filename(self) -> str:
        """
        Get the name of the file.

        Returns:
            str: the file name.
        """
        return self.file.name

    def __enter__(self) -> object:
        """Enter the `with` block."""
        return self

    def __exit__(self, type, value, traceback) -> None:
        """Exit the `with` block."""
        try:
            os.remove(self.file.name)
        except Exception:
            warnings.warn(f'File {self.file.name} could not be closed.')
            self.file.close()



"""Implementation of AnnotatedDataset class."""
import pandas as pd
import torch

from pytoda.warnings import device_warning

from ..types import AnnotatedData, Hashable, List, Union
from .base_dataset import AnyBaseDataset
from .dataframe_dataset import DataFrameDataset


class AnnotatedDataset(DataFrameDataset):
    """
    Annotated samples in order of annotations csv, fetching data
    from passed dataset.
    """

    def __init__(
        self,
        annotations_filepath: str,
        dataset: AnyBaseDataset,
        annotation_index: Union[int, str] = -1,
        label_columns: Union[List[int], List[str]] = None,
        dtype: torch.dtype = torch.float,
        device: torch.device = None,
        **kwargs,
    ) -> None:
        """
        Initialize an annotated dataset via additional annotations dataframe.
        E.g. the  dataset could be SMILES and the annotations could be
        single or multi task labels.

        Args:
            annotations_filepath (str): path to the annotations of a dataset.
                Currently, the supported formats are column separated files.
                The default structure assumes that the last column contains an
                id that is also used in the dataset provided.
            dataset (AnyBaseDataset): instance of a AnyBaseDataset (supporting
                key lookup API of KeyDataset), e.g. a SMILESDataset.
            annotation_index (Union[int, str]): positional or string for the
                column containing the annotation index of keys to get items in
                the passed dataset. Defaults to -1, i.e. the last column.
            label_columns (Union[List[int], List[str]]): indexes (positional
                or strings) for the annotations. Defaults to None, a.k.a. all
                the columns, except the annotation index, are considered
                annotation labels.
            dtype (torch.dtype): torch data type for labels. Defaults to
                torch.float.
            device (torch.device): DEPRECATED
            kwargs (dict): additional parameter for pd.read_csv.
        """
        self.annotations_filepath = annotations_filepath
        self.datasource = dataset
        self.dtype = dtype
        device_warning(device)

        # processing of the dataframe for dataset setup
        df = pd.read_csv(self.annotations_filepath, **kwargs)
        columns = df.columns
        # handle annotation index
        if isinstance(annotation_index, int):
            self.annotation_index = columns[annotation_index]
        elif isinstance(annotation_index, str):
            self.annotation_index = annotation_index
        else:
            raise RuntimeError('annotation_index should be int or str.')

        # handle labels
        if label_columns is None:
            self.labels = [
                column for column in columns if column != self.annotation_index
            ]
        elif all([isinstance(column, int) for column in label_columns]):
            self.labels = columns[label_columns]
        elif all([isinstance(column, str) for column in label_columns]):
            self.labels = label_columns
        else:
            raise RuntimeError(
                'label_columns should be an iterable containing int or str'
            )
        # get the number of labels
        self.number_of_tasks = len(self.labels)

        # set the index explicitly, and discard non label columns
        df = df.set_index(self.annotation_index)[self.labels]
        DataFrameDataset.__init__(self, df)

    def __getitem__(self, index: int) -> AnnotatedData:
        """
        Get key from integer index.

        Args:
            index (int): index annotations to get key and annotation labels,
                where the key is used to fetch the item.

        Returns:
            AnnotatedData: a tuple containing the item itself (with type
                depending on passed dataset) and a torch.Tensor of
                labels for the current item.
        """
        # sample selection
        selected_sample = self.df.iloc[index]
        return self._make_return_tuple(selected_sample)

    def get_item_from_key(self, key: Hashable) -> AnnotatedData:
        """
        Get item via key.

        Args:
            key (Hashable): key of the item and annotations to fetch.

        Returns:
            AnnotatedData: a tuple containing the item itself (with type
                depending on passed dataset) and a torch.Tensor of
                labels for the current item.
        """
        # sample selection
        selected_sample = self.df.loc[key, :]
        return self._make_return_tuple(selected_sample)

    def _make_return_tuple(self, lables_series: pd.Series) -> AnnotatedData:
        # sample
        sample = self.datasource.get_item_from_key(lables_series.name)
        # label
        labels_tensor = torch.tensor(list(lables_series.values), dtype=self.dtype)
        return sample, labels_tensor



"""Testing DrugSensitivityDoseDataset."""
import os
import unittest

import numpy as np
from torch.utils.data import DataLoader

from pytoda.datasets import DrugSensitivityDoseDataset
from pytoda.smiles import SMILESTokenizer
from pytoda.tests.utils import TestFileContent

COLUMN_NAMES = [',drug,cell_line,concentration,viability', ',molecule,omic,dose,label']
DRUG_SENSITIVITY_CONTENT = os.linesep.join(
    [
        '0,CHEMBL14688,sample_3,100,0.7',
        '1,CHEMBL14688,sample_2,23,0.99',
        '2,CHEMBL17564,sample_1,0.4,0.1',
        '3,CHEMBL17564,sample_2,2.3,0.8',
    ]
)
SMILES_CONTENT = os.linesep.join(
    ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
)
GENE_EXPRESSION_CONTENT = os.linesep.join(
    [
        'genes,A,B,C,D',
        'sample_3,9.45,4.984,7.016,8.336',
        'sample_2,7.188,0.695,10.34,6.047',
        'sample_1,9.25,6.133,5.047,5.6',
    ]
)


class TestDrugSensitivityDoseDataset(unittest.TestCase):
    """Testing DrugSensitivityDoseDataset with eager backend."""

    def setUp(self):

        self.smiles_content = SMILES_CONTENT
        self.gene_expression_content = GENE_EXPRESSION_CONTENT

        for column_names in COLUMN_NAMES:
            self.drug_sensitivity_content = os.linesep.join(
                [column_names, DRUG_SENSITIVITY_CONTENT]
            )

            with TestFileContent(
                self.drug_sensitivity_content
            ) as drug_sensitivity_file:
                with TestFileContent(self.smiles_content) as smiles_file:

                    self.smiles_tokenizer = SMILESTokenizer(**{})
                    self.smiles_tokenizer.add_smi(smiles_file.filename)
                    self.smiles_tokenizer.set_max_padding()
                    self.smiles_tokenizer.set_encoding_transforms(
                        padding=True,
                        padding_length=self.smiles_tokenizer.max_token_sequence_length,
                    )
                    with TestFileContent(
                        self.gene_expression_content
                    ) as gene_expression_file:
                        for f in [np.log10, lambda x: x]:
                            self.f = f
                            self.drug_sensitivity_dataset = DrugSensitivityDoseDataset(
                                drug_sensitivity_file.filename,
                                smiles_file.filename,
                                gene_expression_file.filename,
                                smiles_language=self.smiles_tokenizer,
                                gene_expression_kwargs={'pandas_dtype': {'genes': str}},
                                column_names=column_names.split(',')[1:],
                                dose_transform=f,
                            )
                        self.test___len__()
                        self.test___getitem__()
                        # self.test_data_loader()

    def test___len__(self) -> None:
        """Test __len__."""
        self.assertEqual(len(self.drug_sensitivity_dataset), 4)

    def test___getitem__(self) -> None:
        """Test __getitem__."""

        padding_index = self.smiles_tokenizer.padding_index
        c_index = self.smiles_tokenizer.token_to_index['C']
        o_index = self.smiles_tokenizer.token_to_index['O']
        (
            token_indexes_tensor,
            gene_expression_tensor,
            concentration_tensor,
            viability_tensor,
        ) = self.drug_sensitivity_dataset[0]
        np.testing.assert_almost_equal(
            token_indexes_tensor.numpy(),
            np.array([padding_index, padding_index, c_index, o_index]),
        )
        np.testing.assert_almost_equal(
            gene_expression_tensor.numpy(),
            self.drug_sensitivity_dataset.gene_expression_dataset.get_item_from_key(
                'sample_3'
            ).numpy(),
        )
        np.testing.assert_almost_equal(viability_tensor.numpy(), np.array([0.7]))
        np.testing.assert_almost_equal(
            concentration_tensor.numpy(), np.array(self.f([100]))
        )

    def test_data_loader(self) -> None:
        """Test data_loader."""

        data_loader = DataLoader(
            self.drug_sensitivity_dataset, batch_size=3, shuffle=True, drop_last=True
        )
        for (
            batch_index,
            (smiles_batch, gene_expression_batch, conc_batch, viability_batch),
        ) in enumerate(data_loader):
            self.assertEqual(smiles_batch.size(), (3, 4))
            self.assertEqual(gene_expression_batch.size(), (3, 4))
            self.assertEqual(conc_batch.size(), (3, 1))
            self.assertEqual(viability_batch.size(), (3, 1))


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Tuple, Union

import torch
from torch.random import fork_rng
from torch.utils.data import Dataset

from pytoda.types import Any, Tensor
from pytoda.warnings import device_warning


class StochasticItems:
    """Sample an item from a distribution on the fly on indexing.

    Args:
        distribution (torch.distributions.distribution.Distribution): An instance
            of the torch distribution class to sample from. For example, for
            loc = torch.tensor(0.0), and scale=torch.tensor(1.0),
            torch.distributions.normal.Normal(loc,scale), so that
            calling .sample() would return an item from this distribution.
        shape (torch.Size): The desired shape of each item.
        device (torch.device): DEPRECATED
    """

    def __init__(
        self,
        distribution: torch.distributions.distribution.Distribution,
        shape: Union[torch.Size, Tuple[int]],
        device: torch.device = None,
    ):

        self.distribution = distribution
        self.shape = shape
        device_warning(device)

    def __getitem__(self, index: Any) -> Tensor:
        """Samples an item.

        Args:
            index (Any): is ignored.

        Returns:
            Tensor: sampled from distribution with given shape.
        """

        return self.distribution.sample(self.shape)


class DistributionalDataset(Dataset):
    """Generates samples from a specified distribution."""

    def __init__(
        self,
        dataset_size: int,
        item_shape: Tuple[int],
        distribution_function: torch.distributions.distribution.Distribution,
        seed: Optional[int] = None,
        device: torch.device = None,
    ) -> None:
        """Dataset of synthetic samples from a specified distribution with given shape.

        Args:
            dataset_size (int): Number of items to generate (N).
            item_shape (Tuple[int]): The shape of each item tensor returned on
                indexing the dataset. For example for 2D items with timeseries of
                3 timesteps and 5 features: (3, 5)
            distribution_function (torch.distributions.distribution.Distribution):
                An instance of the distribution class from which individual data
                items can be sampled by calling the .sample() method.
                This can either be an object that is directly initialised using a method
                from torch.distributions, such as, torch.distributions.normal.Normal(loc=0.0,scale=1.0),
                or from a factory using a keyword, for example,
                DISTRIBUTION_FUNCTION_FACTORY['normal](loc=0.0, scale=1.0)
                is a valid argument since the factory (found in utils.factories.py)
                initialises the distribution class object based on a string keyword
                and passes the relevant arguments to that object.
            seed (Optional[int]): If passed, all items are generated once with
                this seed (using a local RNG only). Defaults to None, where
                individual items are generated when the DistributionalDataset is
                indexed (using the global RNG).
            device (torch.device): DEPRECATED
        """

        super(DistributionalDataset, self).__init__()

        self.dataset_size = dataset_size
        self.item_shape = item_shape
        self.seed = seed
        device_warning(device)
        self.data_sampler = distribution_function

        if self.seed:
            # Eager dataset creation
            with fork_rng():
                torch.manual_seed(seed)
                torch.cuda.manual_seed(seed)
                torch.cuda.manual_seed_all(seed)

                self.datasource = self.data_sampler.sample((dataset_size, *item_shape))

        else:
            # get sampled item on indexing
            self.datasource = StochasticItems(self.data_sampler, self.item_shape)

    def __len__(self) -> int:
        """Gets length of dataset.

        Returns:
            int: Length of the set being sampled.
        """
        return self.dataset_size

    def __getitem__(self, index: int) -> torch.Tensor:
        """Generates a single set sample.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            torch.Tensor: Tensor containing elements with shape T x H, i.e.
                [self.dataset_depth, self.dataset_dim].
        """

        return self.datasource[index]



"""Processing utilities for .smi files."""
import os
import sys
from typing import List

import pandas as pd
from rdkit import Chem
from rdkit.Chem import Draw

from ..files import read_smi
from ..smiles.transforms import Canonicalization


def filter_invalid_smi(
    input_filepath: str, output_filepath: str, chunk_size: int = 100000
):
    """
    Execute chunked invalid SMILES filtering in a .smi file.

    Args:
        input_filepath (str): path to the .smi file to process.
        output_filepath (str): path where to store the filtered .smi file.
        chunk_size (int): size of the SMILES chunk. Defaults to 100000.
    """
    for chunk in read_smi(input_filepath, chunk_size=chunk_size):
        pd.DataFrame(
            [
                [row['SMILES'], index]
                for index, row in chunk.iterrows()
                if Chem.MolFromSmiles(row['SMILES'])
            ]
        ).to_csv(output_filepath, index=None, header=None, mode='a', sep='\t')


def find_undesired_smiles_files(
    undesired_filepath: str,
    data_filepath: str,
    save_matches: bool = False,
    file=sys.stdout,
    **smi_kwargs,
):
    """
    Method to find undesired SMILES in a list of existing SMILES.

    Args:
        undesired_filepath (str): Path to .smi file with a header at first row.
        data_filepath (str): Path to .csv file with a column 'SMILES'.
        save_matches (bool, optional): Whether found matches should be plotted and
            saved. Defaults to False.
    """
    canonicalizer = Canonicalization()

    # Read undesired data
    undesired = read_smi(undesired_filepath, **smi_kwargs)
    undesired_smiles = undesired['SMILES'].apply(canonicalizer).tolist()

    # Read data filepath
    df = pd.read_csv(data_filepath)

    matches, idxs = [], []
    for idx, row in df.iterrows():
        match = find_undesired_smiles(row['SMILES'], undesired_smiles, canonical=True)

        if match:
            print(f'Found {row.SMILES} in list of undesired SMILES.', file=file)
            matches.append(row.SMILES)
            idxs.append(idx)

    if len(matches) == 0:
        print('No matches found, shutting down.', file=file)
        return

    if save_matches:
        grid = Draw.MolsToGridImage(
            [Chem.MolFromSmiles(s) for s in matches],
            molsPerRow=5,
            maxMols=50,
            legends=[f'Idx: {i}, {s}' for i, s in zip(idxs, matches)],
        )
        grid.save(
            os.path.join(os.path.dirname(data_filepath), 'undesired_molecules.pdf')
        )

    return


def find_undesired_smiles(
    smiles: str, undesired_smiles: List, canonical: bool = False
) -> bool:
    """
    Whether or not a given SMILES is contained in a list of SMILES, respecting
    canonicalization.

    Args:
        smiles (str): Seed SMILES.
        undesired_smiles (List): List of SMILES for comparison
        canonical (bool, optional): Whether comparison list was canonicalized.
            Defaults to False.

    Returns:
        bool: Whether SMILES was present in undesired_smiles.
    """

    canonicalizer = Canonicalization()
    if not canonical:
        undesired_smiles = list(map(canonicalizer, undesired_smiles))

    smiles = canonicalizer(smiles)

    return smiles in undesired_smiles



"""Data splitting utilties."""
import hashlib
import os
import random
from math import ceil
from typing import Tuple

import numpy as np
import pandas as pd

from .types import Files


def csv_data_splitter(
    data_filepaths: Files,
    save_path: str,
    data_type: str,
    mode: str,
    seed: int = 42,
    test_fraction: float = 0.1,
    number_of_columns: int = 12,
    **kwargs,
) -> Tuple[str, str]:
    """
    Function for generic splitting into train and test data in csv format.
    This is an eager splitter trying to fit the entire dataset into memory.

    Args:
        data_filepaths (Files): a list of .csv files that contain the data.
        save_path (str): folder to store the training/testing dataset.
        data_type (str): data type (only used as prefix for the saved files).
        mode (str): mode to split data from: "random" and "file".
            - random: does a random split across all samples in all files.
            - file: randomly splits the files into training and testing.
        seed (int): random seed used for the split. Defaults to 42.
        test_fraction (float): portion of samples in testing data.
            Defaults to 0.1.
        number_of_columns (int): number of columns used to create the hash.
            Defaults to 12.
        kwargs (dict): additional parameters for pd.read_csv.
    Returns:
        Tuple[str, str]: a tuple pointing to the train and test files.
    """
    # preparation
    random.seed(seed)
    data_filepaths = (
        [data_filepaths] if isinstance(data_filepaths, str) else data_filepaths
    )
    file_suffix = '{}_{}_fraction_{}_id_{}_seed_{}.csv'
    hash_fn = hashlib.md5()
    if not ('index_col' in kwargs):
        kwargs['index_col'] = 0

    # helper function to produce unique hash. Base hash on df.to_string()
    # but only consider first number_of_columns columns (slow otherwise)
    def _hash_from_df_columns(df, number_of_columns):
        return df.reindex(sorted(df.columns), axis=1).to_string(
            columns=sorted(df.columns)[:number_of_columns]
        )

    # NOTE: if all *.csv files contain a single sample only
    # the splitting modes collapse
    # file splitting mode
    if mode == 'file':
        if len(data_filepaths) < 2:
            raise RuntimeError(
                'mode={} requires at least two input files.'.format(mode)
            )
        file_indexes = np.arange(len(data_filepaths))
        random.shuffle(file_indexes)
        # compile splits (ceil ensures test_df is not empty)
        test_df = pd.concat(
            [
                pd.read_csv(data_filepaths[index], **kwargs)
                for index in file_indexes[: ceil(test_fraction * len(data_filepaths))]
            ]
        )
        train_df = pd.concat(
            [
                pd.read_csv(data_filepaths[index], **kwargs)
                for index in file_indexes[ceil(test_fraction * len(data_filepaths)) :]
            ]
        )
    # random splitting mode:
    # build a joint df from all samples, then split
    elif mode == 'random':
        df = pd.concat(
            [pd.read_csv(data_path, **kwargs) for data_path in data_filepaths],
            sort=False,
        )
        sample_indexes = np.arange(df.shape[0])
        random.shuffle(sample_indexes)
        splitting_index = ceil(test_fraction * df.shape[0])
        test_df = df.iloc[sample_indexes[:splitting_index]]
        train_df = df.iloc[sample_indexes[splitting_index:]]
    else:
        raise ValueError('Choose mode to be from the set {"random", "file"}.')
    # generate hash (per default uses first number_of_columns
    # columns to define hash)
    number_of_columns = (
        number_of_columns if test_df.shape[1] > number_of_columns else test_df.shape[1]
    )
    hash_str = _hash_from_df_columns(
        train_df, number_of_columns
    ) + _hash_from_df_columns(test_df, number_of_columns)
    hash_fn.update(hash_str.encode('utf-8'))
    hash_id = str(int(hash_fn.hexdigest(), 16))[:6]
    # generate file path
    train_filepath = os.path.join(
        save_path,
        file_suffix.format(data_type, 'train', 1 - test_fraction, hash_id, seed),
    )
    test_filepath = os.path.join(
        save_path, file_suffix.format(data_type, 'test', test_fraction, hash_id, seed)
    )
    # write the dataset
    for path, df in zip([train_filepath, test_filepath], [train_df, test_df]):
        df.to_csv(path)
    return train_filepath, test_filepath



"""Testing Utils."""
import unittest

from pytoda.proteins import aas_to_smiles


class TestUtils(unittest.TestCase):
    """Testing Utils."""

    def test_aas_to_smiles(self) -> None:
        """Test aas_to_smiles."""
        sequences = ['EGK', 'A']
        ground_truth = [
            'NCCCC[C@H](NC(=O)CNC(=O)[C@@H](N)CCC(=O)O)C(=O)O',
            'C[C@H](N)C(=O)O',
        ]

        for seq, gt in zip(sequences, ground_truth):
            self.assertEqual(gt, aas_to_smiles(seq))


if __name__ == '__main__':
    unittest.main()



"""Testing SMILESLanguage."""
import os
import tempfile
import unittest

import torch

from pytoda.smiles.processing import split_selfies
from pytoda.smiles.smiles_language import (
    SELFIESLanguage,
    SMILESLanguage,
    SMILESTokenizer,
)
from pytoda.smiles.transforms import Selfies
from pytoda.tests.utils import TestFileContent


class TestSmilesLanguage(unittest.TestCase):
    """Testing SMILESLanguage."""

    def test__update_max_token_sequence_length(self) -> None:
        """Test _update_max_token_sequence_length."""
        smiles = 'CCO'
        smiles_language = SMILESLanguage()
        self.assertEqual(smiles_language.max_token_sequence_length, 0)
        smiles_language.add_smiles(smiles)
        self.assertEqual(smiles_language.max_token_sequence_length, 3)
        smiles_language = SMILESTokenizer(add_start_and_stop=True)
        self.assertEqual(smiles_language.max_token_sequence_length, 0)
        smiles_language.add_smiles(smiles)
        self.assertEqual(smiles_language.max_token_sequence_length, 5)

    def test__update_language_dictionaries_with_tokens(self) -> None:
        """Test _update_language_dictionaries_with_tokens."""
        smiles = 'CCO'
        smiles_language = SMILESLanguage()
        smiles_language._update_language_dictionaries_with_tokens(
            smiles_language.smiles_tokenizer(smiles)
        )
        self.assertTrue(
            'C' in smiles_language.token_to_index
            and 'O' in smiles_language.token_to_index
        )
        self.assertEqual(smiles_language.number_of_tokens, 37)

    def test_add_smis(self) -> None:
        """Test add_smis."""
        content = os.linesep.join(
            ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
        )
        with TestFileContent(content) as a_test_file:
            with TestFileContent(content) as another_test_file:
                smiles_language = SMILESLanguage()
                smiles_language.add_smis(
                    [a_test_file.filename, another_test_file.filename]
                )
                self.assertEqual(smiles_language.number_of_tokens, 39)

    def test_add_smi(self) -> None:
        """Test add_smi."""
        content = os.linesep.join(
            ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
        )
        with TestFileContent(content) as test_file:
            smiles_language = SMILESLanguage()
            smiles_language.add_smi(test_file.filename)
            self.assertEqual(smiles_language.number_of_tokens, 39)

    def test_add_smiles(self) -> None:
        """Test add_smiles."""
        smiles = 'CCO'
        smiles_language = SMILESLanguage()
        smiles_language.add_smiles(smiles)
        self.assertEqual(smiles_language.number_of_tokens, 37)

    def test_add_token(self) -> None:
        """Test add_token."""
        token = 'token'
        smiles_language = SMILESLanguage()
        smiles_language.add_token(token)
        self.assertEqual(smiles_language.number_of_tokens, 36)

    def test_multi_processing(self) -> None:
        """Test multiprocessing"""

        for start_stop in [False, True]:
            s = SMILESTokenizer(add_start_and_stop=start_stop)
            try:
                torch.multiprocessing.spawn(
                    s, args=(), nprocs=1, join=True, daemon=False, start_method='spawn'
                )
            except AttributeError:
                # Catch this exception since it is likely caused by Pickling Error
                raise AttributeError()
            except Exception:
                # Other errors are okay since s is not callable and raises a TypeError
                # But spawn doesn't allow to catch error types, so we use bare Exception
                pass

    def test_smiles_to_token_indexes(self) -> None:
        """Test smiles_to_token_indexes."""
        smiles = 'CCO'
        smiles_language = SMILESLanguage()
        smiles_language.add_smiles(smiles)
        token_indexes = [smiles_language.token_to_index[token] for token in smiles]
        self.assertListEqual(
            smiles_language.smiles_to_token_indexes(smiles), token_indexes
        )

        # unknown token
        smiles_u = 'CCN'
        token_indexes_u = [
            smiles_language.token_to_index['C'],
            smiles_language.token_to_index['C'],
            smiles_language.unknown_index,
        ]
        self.assertListEqual(
            smiles_language.smiles_to_token_indexes(smiles_u), token_indexes_u
        )

        smiles_language = SMILESTokenizer(add_start_and_stop=True, padding=True)
        # fail with no padding_length defined
        with self.assertRaises(TypeError):
            smiles_language.smiles_to_token_indexes(smiles)

        smiles_language.add_smiles(smiles)
        smiles_language.set_max_padding()
        self.assertListEqual(
            list(smiles_language.smiles_to_token_indexes(smiles)),
            [smiles_language.start_index]
            + token_indexes
            + [smiles_language.stop_index],
        )

        # SELFIES
        smiles_language = SMILESLanguage(smiles_tokenizer=split_selfies)
        transform = Selfies()
        selfies = transform(smiles)
        self.assertEqual(selfies, smiles_language.smiles_to_selfies(smiles))
        smiles_language.add_smiles(selfies)
        token_indexes = [
            smiles_language.token_to_index[token] for token in ['[C]', '[C]', '[O]']
        ]
        self.assertListEqual(
            smiles_language.smiles_to_token_indexes(selfies), token_indexes
        )
        smiles_language = SMILESTokenizer(
            add_start_and_stop=True, smiles_tokenizer=split_selfies
        )
        smiles_language.add_smiles(selfies)
        self.assertListEqual(
            list(smiles_language.smiles_to_token_indexes(selfies)),
            [smiles_language.start_index]
            + token_indexes
            + [smiles_language.stop_index],
        )

    def test_token_indexes_to_smiles(self) -> None:
        """Test token_indexes_to_smiles."""
        smiles = 'CCO'
        smiles_language = SMILESLanguage()
        smiles_language.add_smiles(smiles)
        token_indexes = [smiles_language.token_to_index[token] for token in smiles]
        self.assertEqual(smiles_language.token_indexes_to_smiles(token_indexes), 'CCO')
        token_indexes = (
            [smiles_language.start_index] + token_indexes + [smiles_language.stop_index]
        )
        smiles_language = SMILESTokenizer(add_start_and_stop=True)
        smiles_language.add_smiles(smiles)
        self.assertEqual(smiles_language.token_indexes_to_smiles(token_indexes), 'CCO')

    def test_smiles_roundtrip(self) -> None:
        """Test smiles_to_token_indexes and token_indexes_to_smiles."""
        smiles = 'CCO'
        smiles_language = SMILESLanguage()
        smiles_language.add_smiles(smiles)

        self.assertEqual(
            smiles_language.token_indexes_to_smiles(
                smiles_language.smiles_to_token_indexes('CCO')
            ),
            'CCO',
        )

        smiles_language = SMILESTokenizer(add_start_and_stop=True)
        smiles_language.add_smiles(smiles)
        self.assertEqual(
            smiles_language.token_indexes_to_smiles(
                smiles_language.smiles_to_token_indexes('CCO')
            ),
            'CCO',
        )

    def test_smiles_to_selfies(self) -> None:
        """Test smiles_to_selfies and selfies_to_smiles"""
        smiles = 'CCO'
        smiles_language = SMILESLanguage()

        self.assertEqual(
            smiles,
            smiles_language.selfies_to_smiles(
                smiles_language.smiles_to_selfies(smiles)
            ),
        )

    def test_transform_changes(self):
        """Test update of transforms on parameter assignment"""
        smiles = 'CCO'
        smiles_language = SMILESTokenizer()
        self.assertEqual(smiles_language.max_token_sequence_length, 0)

        smiles_language.add_smiles(smiles)
        self.assertEqual(smiles_language.max_token_sequence_length, 3)

        smiles_language.add_start_and_stop = True
        smiles_language.add_smiles(smiles)  # is length function set?
        self.assertEqual(smiles_language.max_token_sequence_length, 5)
        tokens = smiles_language.smiles_to_token_indexes(smiles)

        smiles_language.padding = True
        print("\nExpected warning on setting padding length too small:")
        smiles_language.padding_length = 2  # 2 < 5
        print("\nExpected warning when actually truncating token indexes:")
        tokens_ = smiles_language.smiles_to_token_indexes(smiles)

        self.assertEqual(len(tokens_), 2)
        smiles_language.padding = False
        self.assertSequenceEqual(
            list(tokens), list(smiles_language.smiles_to_token_indexes(smiles))
        )

        smiles_language.set_encoding_transforms(padding=True, padding_length=2)
        print("\nExpected warning when actually truncating token indexes:")
        self.assertSequenceEqual(
            list(tokens_), list(smiles_language.smiles_to_token_indexes(smiles))
        )

        smiles_language.reset_initial_transforms()
        self.assertSequenceEqual(
            list(tokens), list(smiles_language.smiles_to_token_indexes(smiles))
        )

    def test_vocab_roundtrip(self):
        smiles = 'CCO'
        source_language = SMILESLanguage()
        source_language.add_smiles(smiles)
        # to test
        vocab = source_language.token_to_index
        vocab_ = source_language.index_to_token
        max_len = source_language.max_token_sequence_length
        count = source_language.token_count
        total = source_language.number_of_tokens

        # just vocab
        with tempfile.TemporaryDirectory() as tempdir:
            source_language.save_vocabulary(tempdir)

            smiles_language = SMILESTokenizer()
            smiles_language.load_vocabulary(tempdir)
        self.assertDictEqual(vocab, smiles_language.token_to_index)
        self.assertDictEqual(vocab_, smiles_language.index_to_token)

        # pretrained
        with tempfile.TemporaryDirectory() as tempdir:
            source_language.save_pretrained(tempdir)

            smiles_language = SMILESTokenizer.from_pretrained(tempdir)
        self.assertDictEqual(vocab, smiles_language.token_to_index)
        self.assertDictEqual(vocab_, smiles_language.index_to_token)

        self.assertEqual(max_len, smiles_language.max_token_sequence_length)
        self.assertDictEqual(count, smiles_language.token_count)
        self.assertEqual(total, smiles_language.number_of_tokens)


class TestSelfiesLanguage(unittest.TestCase):
    """Testing SELFIESLanguage."""

    def test_add_dataset(self):
        dataset = ['C(', 'CC)', 'CCO', 'C', 'CO', 'NCCS']
        selfies_language = SELFIESLanguage()
        print("\nExpected 'SMILES Parse Error's while iterating a dataset:")
        selfies_language.add_dataset(dataset)
        self.assertTrue((0, 'C(') in selfies_language.invalid_molecules)
        self.assertTrue((1, 'CC)') in selfies_language.invalid_molecules)


if __name__ == '__main__':
    unittest.main()



"""PolymerTokenizerDataset module."""
import pandas as pd
import torch
from numpy import iterable
from torch.utils.data import Dataset

from pytoda.warnings import device_warning

from ..smiles.polymer_language import PolymerTokenizer
from ..types import List, Sequence, Tensor, Tuple, Union
from .smiles_dataset import SMILESDataset


class PolymerTokenizerDataset(Dataset):
    """
    Dataset of SMILES from multiple entities encoded as token indexes.

    Creates a tuple of SMILES datasets, one per given entity (i.e. molecule
    class, e.g monomer and catalysts).
    Rows in the annotation df needs to have column names identical to entities,
    mapping to SMILES in the datasets.

    Uses a PolymerTokenizer
    """

    def __init__(
        self,
        *smi_filepaths: str,
        entity_names: Sequence[str],
        annotations_filepath: str,
        annotations_column_names: Union[List[int], List[str]] = None,
        smiles_language: PolymerTokenizer = None,
        canonical: Union[Sequence[bool], bool] = False,
        augment: Union[Sequence[bool], bool] = False,
        kekulize: Union[Sequence[bool], bool] = False,
        all_bonds_explicit: Union[Sequence[bool], bool] = False,
        all_hs_explicit: Union[Sequence[bool], bool] = False,
        randomize: Union[Sequence[bool], bool] = False,
        remove_bonddir: Union[Sequence[bool], bool] = False,
        remove_chirality: Union[Sequence[bool], bool] = False,
        selfies: Union[Sequence[bool], bool] = False,
        sanitize: Union[Sequence[bool], bool] = True,
        padding: Union[Sequence[bool], bool] = True,
        padding_length: Union[Sequence[int], int] = None,
        iterate_dataset: bool = True,
        backend: str = 'eager',
        device: torch.device = None,
        **kwargs,
    ) -> None:
        """
        Initialize a Polymer dataset.

        All SMILES dataset parameter can be controlled either separately for
        each dataset (by iterable of correct length) or globally (bool/int).

        Args:
            smi_filepaths (Files): paths to .smi files, one per entity
            entity_names (Sequence[str]): List of chemical entities.
            annotations_filepath (str): Path to .csv with the IDs of the
                chemical entities and their properties. Needs to have one
                column per entity name.
            annotations_column_names (Union[List[int], List[str]]): indexes
                (positional or strings) for the annotations. Defaults to None,
                a.k.a. all the columns, except the entity_names are annotation
                labels.
            smiles_language (PolymerTokenizer): a polymer language.
                Defaults to None, in which case a new object is created.
            padding (Union[Sequence[bool], bool]): pad sequences to longest in
                the smiles language. Defaults to True. Controlled either for
                each dataset separately (by iterable) or globally (bool).
            padding_length (Union[Sequence[int], int]): manually sets number of
                applied paddings, applies only if padding is True. Defaults to
                None. Controlled either for each dataset separately (by
                iterable) or globally (int).
            canonical (Union[Sequence[bool], bool]): performs canonicalization
                of SMILES (one original string for one molecule), if True, then
                other transformations (augment etc, see below) do not apply.
            augment (Union[Sequence[bool], bool]): perform SMILES augmentation.
                Defaults to False.
            kekulize (Union[Sequence[bool], bool]): kekulizes SMILES
                (implicit aromaticity only).
                Defaults to False.
            all_bonds_explicit (Union[Sequence[bool], bool]): Makes all bonds
                explicit. Defaults to False, only applies if kekulize = True.
            all_hs_explicit (Union[Sequence[bool], bool]): Makes all hydrogens
                explicit. Defaults to False, only applies if kekulize = True.
            randomize (Union[Sequence[bool], bool]): perform a true
                randomization of SMILES tokens. Defaults to False.
            remove_bonddir (Union[Sequence[bool], bool]): Remove directional
                info of bonds. Defaults to False.
            remove_chirality (Union[Sequence[bool], bool]): Remove chirality
                information. Defaults to False.
            selfies (Union[Sequence[bool], bool]): Whether selfies is used
                instead of smiles, defaults to False.
            sanitize (Union[Sequence[bool], bool]): sanitize (bool): RDKit
                sanitization of the molecule. Defaults to True.
            iterate_dataset (bool): whether to go through all SMILES in the
                dataset to build/extend vocab, find longest sequence, and
                checks the passed padding length if applicable. Defaults to
                True.
            backend (str): memory management backend.
                Defaults to eager, prefer speed over memory consumption.
            device (torch.device): DEPRECATED
            kwargs (dict): additional arguments for dataset constructor.

        NOTE: If a parameter that can be given as Union[Sequence[bool], bool]
        is given as Sequence[bool] of wrong length (!= len(entity_names)), the
        first list item is used for all datasets.
        """
        device_warning(device)
        self.backend = backend

        if len(entity_names) != len(smi_filepaths):
            raise ValueError('Give 1 .smi file per entity')

        # Setup parameter

        (
            self.paddings,
            self.padding_lengths,
            self.canonicals,
            self.augments,
            self.kekulizes,
            self.all_bonds_explicits,
            self.all_hs_explicits,
            self.randomizes,
            self.remove_bonddirs,
            self.remove_chiralitys,
            self.selfies,
            self.sanitize,
        ) = map(
            (
                lambda x: x
                if iterable(x) and len(x) == len(entity_names)
                else [x] * len(entity_names)
            ),
            (
                padding,
                padding_length,
                canonical,
                augment,
                kekulize,
                all_bonds_explicit,
                all_hs_explicit,
                randomize,
                remove_bonddir,
                remove_chirality,
                selfies,
                sanitize,
            ),
        )

        if smiles_language is None:
            self.smiles_language = PolymerTokenizer(  # defaults to add smiles
                entity_names=entity_names,
                padding=self.paddings[0],
                padding_length=self.padding_lengths[0],
                canonical=self.canonicals[0],
                augment=self.augments[0],
                kekulize=self.kekulizes[0],
                all_bonds_explicit=self.all_bonds_explicits[0],
                all_hs_explicit=self.all_hs_explicits[0],
                randomize=self.randomizes[0],
                remove_bonddir=self.remove_bonddirs[0],
                remove_chirality=self.remove_chiralitys[0],
                selfies=self.selfies[0],
                sanitize=self.sanitize[0],
                add_start_and_stop=True,
            )
            for index, entity in enumerate(entity_names):
                self.smiles_language.set_smiles_transforms(
                    entity,
                    canonical=self.canonicals[index],
                    augment=self.augments[index],
                    kekulize=self.kekulizes[index],
                    all_bonds_explicit=self.all_bonds_explicits[index],
                    all_hs_explicit=self.all_hs_explicits[index],
                    remove_bonddir=self.remove_bonddirs[index],
                    remove_chirality=self.remove_chiralitys[index],
                    selfies=self.selfies[index],
                    sanitize=self.sanitize[index],
                )
                # set_encoding_transforms only after adding smiles
                # while transforms are needed to add_dataset

        else:
            self.smiles_language = smiles_language

        self.entities = self.smiles_language.entities
        self.datasets = [
            SMILESDataset(smi_filepath, name=self.entities[index], **kwargs)
            for index, smi_filepath in enumerate(smi_filepaths)
        ]

        if iterate_dataset:
            for dataset in self.datasets:
                self.smiles_language.update_entity(dataset.name)
                self.smiles_language.add_dataset(dataset)
            if padding and None in self.padding_lengths:
                # take care, this will call a transform reset
                self.smiles_language.set_max_padding()
        self.smiles_language.current_entity = None

        if smiles_language is None:
            for index, entity in enumerate(entity_names):
                # smiles_transforms might have been reset
                self.smiles_language.set_smiles_transforms(
                    entity,
                    canonical=self.canonicals[index],
                    augment=self.augments[index],
                    kekulize=self.kekulizes[index],
                    all_bonds_explicit=self.all_bonds_explicits[index],
                    all_hs_explicit=self.all_hs_explicits[index],
                    remove_bonddir=self.remove_bonddirs[index],
                    remove_chirality=self.remove_chiralitys[index],
                    selfies=self.selfies[index],
                    sanitize=self.sanitize[index],
                )
                self.smiles_language.set_encoding_transforms(
                    entity,
                    randomize=self.randomizes[index],
                    add_start_and_stop=True,
                    padding=self.paddings[index],
                    padding_length=self.padding_lengths[index],
                )

        # Read and post-process the annotations dataframe
        self.annotations_filepath = annotations_filepath
        self.annotated_data_df = pd.read_csv(self.annotations_filepath)
        # Cast the column names to uppercase
        self.annotated_data_df.columns = map(
            lambda x: str(x).capitalize(), self.annotated_data_df.columns
        )
        columns = self.annotated_data_df.columns

        # handle annotation index
        assert all(
            [entity in columns for entity in self.entities]
        ), 'Some of the chemical entities were not found in the label csv.'

        # handle labels
        if annotations_column_names is None:
            self.labels = [column for column in columns if column not in self.entities]
        elif all([isinstance(column, int) for column in annotations_column_names]):
            self.labels = columns[annotations_column_names]
        elif all([isinstance(column, str) for column in annotations_column_names]):
            self.labels = list(map(lambda x: x.capitalize(), annotations_column_names))
        else:
            raise RuntimeError(
                'label_columns should be an iterable containing int or str'
            )
        # get the number of labels
        self.number_of_tasks = len(self.labels)

        # filter data based on the availability
        masks = []
        mask = pd.Series(
            [True] * len(self.annotated_data_df), index=self.annotated_data_df.index
        )
        for entity, dataset in zip(self.entities, self.datasets):
            # prune rows (in mask) with ids unavailable in respective dataset
            local_mask = self.annotated_data_df[entity].isin(set(dataset.keys()))
            mask = mask & local_mask
            masks.append(local_mask)

        self.annotated_data_df = self.annotated_data_df.loc[mask]

        # to investigate missing ids per entity
        self.masks_df = pd.concat(masks, axis=1)
        self.masks_df.columns = self.entities

    def __len__(self) -> int:
        """Total number of samples."""
        return len(self.annotated_data_df)

    def __getitem__(self, index: int) -> Tuple[Tensor, ...]:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            Tuple: a tuple containing self.entities+1 torch.Tensors
            representing respectively: compound token indexes for each chemical
            entity and the property labels (annotations)
        """

        # sample selection for all entities/datasets
        selected_sample = self.annotated_data_df.iloc[index]

        # labels (annotations)
        labels_tensor = torch.tensor(
            list(selected_sample[self.labels].values),
            dtype=torch.float,
        )
        # samples (SMILES token indexes)
        smiles_tensors = tuple(
            self.smiles_language.smiles_to_token_indexes(
                dataset.get_item_from_key(selected_sample[dataset.name]), dataset.name
            )
            for dataset in self.datasets
        )
        return tuple([*smiles_tensors, labels_tensor])



"""SMILES processing utilities."""
import codecs
import logging
import re
import warnings

from importlib_resources import as_file, files
from selfies import split_selfies as split_selfies_
from SmilesPE.pretokenizer import kmer_tokenizer
from SmilesPE.tokenizer import SPE_Tokenizer

from ..types import Dict, Tokenizer, Tokens

logger = logging.getLogger(__name__)

# tokenizer
SMILES_TOKENIZER = re.compile(
    r'(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|'
    r'-|\+|\\\\|\/|:|~|@|\?|>|\*|\$|\%[0-9]{2}|[0-9])'
)

with as_file(files('pytoda.smiles.metadata').joinpath('spe_chembl.txt')) as filepath:
    SPE_TOKENIZER = SPE_Tokenizer(codecs.open(str(filepath)))


def tokenize_smiles(smiles: str, regexp=SMILES_TOKENIZER, *args, **kwargs) -> Tokens:
    """
    Tokenize a character-level SMILES string.

    Args:
        smiles (str): a SMILES representation.
        regexp (re.Pattern): optionally pass a regexp for the
            tokenization. Defaults to SMILES_TOKENIZER.
        args (): ignored, for backwards compatibility.
        kwargs (): ignored, for backwards compatibility.
    Returns:
        Tokens: the tokenized SMILES.
    """
    return [token for token in regexp.split(smiles) if token]


def kmer_smiles_tokenizer(
    smiles: str, k: int = 2, stride: int = 1, *args, **kwargs
) -> Tokens:
    """K-Mer SMILES tokenization following SMILES PE (Li et al. 2020):
        Li, Xinhao, and Denis Fourches. "SMILES Pair Encoding: A Data-Driven
        Substructure Tokenization Algorithm for Deep Learning." (2020).


    Args:
        smiles (str): SMILES string to be tokenized.
        k (int): Positive integer denoting the tuple/k-gram lengths. Defaults
            to 2 (bigrams).
        stride (int, optional): Stride used for k-mer generation. Higher values
            result in less tokens. Defaults to 1 (densely overlapping).
        args (): Optional arguments for `kmer_tokenizer`.
        kwargs (): Optional keyword arguments for `kmer_tokenizer`.

    Returns:
        Tokens: Tokenized SMILES sequence (list of str).
    """

    return kmer_tokenizer(smiles, ngram=k, stride=stride, *args, **kwargs)


def spe_smiles_tokenizer(smiles: str) -> Tokens:
    """Pretrained SMILES Pair Encoding tokenizer following (Li et al. 2020).
        Splits a SMILES into tokens of substructures of varying lengths,
        depending on occurrence of tokens in ChEMBL dataset.

        Li, Xinhao, and Denis Fourches. "SMILES Pair Encoding: A Data-Driven
        Substructure Tokenization Algorithm for Deep Learning." (2020).

    Args:
        smiles (str): SMILES string to be tokenized.

    Returns:
        Tokens: SMILES tokenized into substructures (list of str).
    """

    return SPE_TOKENIZER.tokenize(smiles).split(' ')


def tokenize_selfies(selfies: str) -> Tokens:
    """Tokenize SELFIES.

    NOTE: Code adapted from selfies package (`def selfies_to_hot`):
        https://github.com/aspuru-guzik-group/selfies

    Args:
        selfies (str): a SELFIES representation (character-level).

    Returns:
        Tokens: the tokenized SELFIES.
    """
    warnings.warn(
        "tokenize_selfies will be deprecated in favor of `split_selfies`",
        DeprecationWarning,
    )
    try:
        selfies = selfies.replace('.', '[.]')  # to allow parsing unbound atoms
        selfies_char_list_pre = selfies[1:-1].split('][')
        return [
            '[' + selfies_element + ']' for selfies_element in selfies_char_list_pre
        ]
    except Exception:
        logger.warning(f'Error in tokenizing {selfies}. Returning empty list.')
        return ['']


def split_selfies(selfies: str) -> Tokens:
    """Tokenize SELFIES, wrapping generator as list.

    Args:
        selfies (str): a SELFIES representation (character-level).

    Returns:
        Tokens: the tokenized SELFIES.
    """
    return list(split_selfies_(selfies))


TOKENIZER_FUNCTIONS: Dict[str, Tokenizer] = {
    'smiles': tokenize_smiles,
    'kmer_smiles': kmer_smiles_tokenizer,
    'spe_smiles': spe_smiles_tokenizer,
    'selfies': split_selfies,
}



"""Implementation of _FastaLazyDataset."""
from itertools import islice

from pyfaidx import Fasta

from ..types import Callable, Hashable, Iterator
from .base_dataset import KeyDataset


class _FastaLazyDataset(KeyDataset):
    """
    .fasta dataset using lazy loading via pyfaidx, which creates a small flat
    index file ".fai".

    Suggested when handling fasta without uniprot headers and datasets that
    can't fit in the device memory.
    """

    def __init__(
        self,
        fasta_filepath: str,
        name: str = 'Sequence',
        key_function: Callable[[str], str] = lambda x: x,
        **kwargs,
    ) -> None:
        """Initialize a .fasta dataset. with .fai index file.

        Args:
            fasta_filepath (str): path to .fasta file.
            name (str, optional): type of dataset. Defaults to 'Sequence'.
            key_function (Callable[[str], str], optional): function returning a
                unique key given the FASTA sequence header. Defaults to
                identity.
            kwargs (dict): Additional parameters passed to pyfaidx.Fasta class.

        Raises:
            ValueError: in case of duplicate keys from key_function in dataset.

        """
        self.fasta_filepath = fasta_filepath
        self.name = name
        self.datasource = Fasta(
            filename=fasta_filepath, key_function=key_function, as_raw=False, **kwargs
        )
        KeyDataset.__init__(self)

        self.key_to_index_mapping = {
            key: index for index, key in enumerate(self.datasource.keys())
        }

    def __len__(self) -> int:
        """Total number of samples."""
        return len(self.datasource.records)

    def __getitem__(self, index: int) -> str:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            str: a sequence in the FASTA file.
        """
        # Fasta class indexes via key too, but iterates full keys
        return self.get_item_from_key(self.get_key(index))

    def get_item_from_key(self, key: Hashable) -> str:
        """Get item via sample identifier"""
        return str(self.datasource.records[key])

    def get_key(self, index: int) -> Hashable:
        """Get sample identifier from integer index."""
        return next(islice(self.datasource.keys(), index, None))

    def get_index(self, key: Hashable) -> int:
        """Get index for first datum mapping to the given sample identifier."""
        return self.key_to_index_mapping[key]

    def keys(self) -> Iterator:
        """Default iterator of keys by iterating over dataset indexes."""
        return iter(self.datasource.keys())

    @property
    def has_duplicate_keys(self):
        # would error on init
        return False



from warnings import warn

from .types import Any


def device_warning(device: Any, throw_exception: bool = True):
    """

    Args:
        device: A device object provided by the user.
        throw_exception: Whether an exception is thrown. Defaults to True.

    Raises:
        ValueError: Raises if throw_exception is True.

    """

    if device:
        msg = (
            f"Found deprecated argument: device={device}. Device value will be ignored, "
            "all devices will be set to `cpu`. If you are using cuda devices, also "
            "remember to transfer tensors to the gpu before using a model __call__"
        )
        if throw_exception:
            raise ValueError(msg)
        warn(msg)



"""Amino Acid Sequence transforms."""
from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np

from ..files import read_smi
from ..transforms import Transform
from ..types import Indexes
from .processing import REAL_AAS
from .protein_language import ProteinLanguage


class SequenceToTokenIndexes(Transform):
    """Transform Sequence to token indexes using Sequence language."""

    def __init__(self, protein_language: ProteinLanguage) -> None:
        """
        Initialize a Sequence to token indexes object.

        Args:
            protein_language (ProteinLanguage): a Protein language.
        """
        self.protein_language = protein_language

    def __call__(self, smiles: str) -> Indexes:
        """
        Apply the Sequence tokenization transformation

        Args:
            smiles (str): a Sequence representation.

        Returns:
            Indexes: indexes representation for the Sequence provided.
        """
        return self.protein_language.sequence_to_token_indexes(smiles)


class ReplaceByFullProteinSequence(Transform):
    """
    A transform to replace short amino acid sequences with the full protein sequence.
    For example, replace active site sequence of a kinase with its full sequence.
    """

    def __init__(self, alignment_path: Union[str, Path]) -> None:
        """
        Loads alignment info with two "classes" (or types) of residues.

        Args:
            alignment_path (str): path to `.smi` or `.tsv` file which allows to map
                between shortened and full, aligned sequences. Do not use a header in
                the file.

                NOTE: By convention, residues in upper case are important and will be
                kept and residues in lower case are less important and are (usually)
                discarded.
                NOTE: The first column has to be the full protein sequence (use upper
                case only for residues to be used). E.g., ggABCggDEFgg
                NOTE: The second column has to be the condensed sequence (ABCDEF).
                NOTE: The third column has to be a protein id (can be duplicated).
        """

        if not (isinstance(alignment_path, str) or isinstance(alignment_path, Path)):
            raise TypeError(
                f"alignment_path must be string or Path, not {type(alignment_path)}"
            )
        self.alignment_path = alignment_path

        alignment_info = read_smi(
            self.alignment_path,
            index_col=None,
            header=None,
            names=['full_sequence', 'short_sequence', 'id'],
        )
        # We use a combination of ID and the shortened sequence as keys to
        # enable support if proteins have >1 short sequence.
        self.id_to_full = dict(
            zip(
                alignment_info['id'] + '_' + alignment_info['short_sequence'],
                alignment_info['full_sequence'],
            )
        )
        if len(self.id_to_full) < len(alignment_info):
            raise ValueError(f'Duplicate IDs not allowed in: {self.alignment_path}')

    def __call__(self, sample_dict: Dict) -> str:
        """
        Replace the shortened sequence (usually uppercase only) with an aligned
        sequence where usually uppercase is for residues of interest and lowercase
        for the remaining ones.

        Args:
            sample_dict (Dict): a dictionary with the following keys:
                - 'id': the protein id.
                - 'sequence': the shortened protein sequence. (E.g., ABCDEF)
            NOTE: This has to be a dictionary because otherwise the shortened protein
                sequence has to be unique.

        Returns:
            str: the full protein sequence (e.g., abABChijDEF).
        """
        return self.id_to_full[f"{sample_dict['id']}_{sample_dict['sequence']}"]


def extract_active_sites_info(
    aligned_seq: str,
) -> Tuple[str, List[str], List[str], List[str]]:
    """
    Processes and extracts useful information from an aligned protein sequence.
    Expects lower case amino acids to be outside of the relevant area (e.g., active site)
    and upper case amino acids to be inside it.

    Args:
        aligned_seq: A string containing the aligned protein sequence including
            lower case amino acids and high case amino acids.

    Returns:
        4-Tuple of:
        aligned_seq (str): The input sequence.
        non_active_sites (List[str]): A list of strings, one item for each contiguous
            subsequence NOT belonging to active site.
        active_sites (List[str]): A list of strings, one item for each contiguous
            subsequence belonging to active site.
        all_seqs (List[str]): A list of strings, one item for each contiguous
            subsequence that either belongs to the active site or not.
    """

    non_active_sites = ''
    active_sites = ''
    prev_was_highcase = False
    for c in aligned_seq:
        next_is_highcase = c <= 'Z'
        if next_is_highcase ^ prev_was_highcase:
            if next_is_highcase:
                active_sites += '#'
            else:
                non_active_sites += '#'

        if next_is_highcase:
            active_sites += c
            prev_was_highcase = True
        else:
            non_active_sites += c
            prev_was_highcase = False

    non_active_sites = [s for s in non_active_sites.split('#') if s != '']
    active_sites = [s for s in active_sites.split('#') if s != '']

    if aligned_seq[0] <= 'Z':
        zip_obj = zip(active_sites, non_active_sites)
    else:
        zip_obj = zip(non_active_sites, active_sites)

    all_seqs = [i for one_tuple in zip_obj for i in one_tuple]

    if len(active_sites) > len(non_active_sites):
        assert len(active_sites) == len(non_active_sites) + 1
        all_seqs.append(active_sites[-1])
    elif len(active_sites) < len(non_active_sites):
        assert len(active_sites) + 1 == len(non_active_sites)
        all_seqs.append(non_active_sites[-1])

    return aligned_seq, non_active_sites, active_sites, all_seqs


def verify_aligned_info(sequence: str) -> None:
    """
    Verify that the sequence is aligned.

    Args:
        sequence: An amino acid sequence.

    Raises:
        Exception: If alignment could not be detected.
    """
    isinstance(sequence, str)
    found_lower_case = False
    found_upper_case = False
    for c in sequence:
        if c >= 'A' and c <= 'Z':
            found_upper_case = True
        elif c >= 'a' and c <= 'z':
            found_lower_case = True
    if not (found_lower_case and found_upper_case):
        raise Exception(
            'Expected aligned residues sequence! Did you forget to use ReplaceByFullProteinSequence?'
        )


class ProteinAugmentFlipSubstrs(Transform):
    """Augment a protein sequence by randomly flipping each contiguous subsequence."""

    def __init__(self, p: float = 0.5) -> None:
        """
        Args:
            p (float): Probability that reverting occurs.
        """
        if not isinstance(p, float):
            raise TypeError(f'Please pass float, not {type(p)}.')
        self._p = np.clip(p, 0.0, 1.0)

    def __call__(self, sequence: str) -> str:
        """
        Apply the transform.

        Args:
            sequence (str): an aligned sequence (example: abcDEfgHI).

        Returns:
            str: an aligned sequence with optional flipping (example: abcEDfgHI).
        """
        verify_aligned_info(sequence)
        (
            aligned_seq,
            non_active_sites,
            active_sites,
            all_seqs,
        ) = extract_active_sites_info(sequence)

        ans = ''
        for substr in all_seqs:
            if substr[0] <= 'Z':
                if np.random.rand() < self._p:
                    ans += substr[::-1]
                else:
                    ans += substr
            else:
                ans += substr

        return ans


class MutateResidues(Transform):
    """
    Augment a protein sequence by injecting (possibly different) noise to residues
    inside and outside the relevant part (e.g., active site).
    NOTE: Noise means single-residue point mutations.
    """

    def __init__(self, mutate_upper: float = 0.01, mutate_lower: float = 0.1) -> None:
        """
        Args:
            mutate_lower (float): probability for mutating lowercase residues
            mutate_upper (float): probability for mutating uppercase residues.
        """

        if not isinstance(mutate_upper, float):
            raise TypeError(
                f'Please pass float for mutate_prob_in_active_site, not {type(mutate_upper)}.'
            )
        self.mutate_upper = mutate_upper

        if not isinstance(mutate_lower, float):
            raise TypeError(
                f'Please pass float for mutate_lower, not {type(mutate_lower)}.'
            )
        self.mutate_lower = mutate_lower
        self.num_aas = len(REAL_AAS)

    def __call__(self, sequence: str) -> str:
        """
        Apply the transform.

        Args:
            sequence (str): an aligned sequence (example: acDEFg).

        Returns:
            str: a possibly mutated aligned sequence (example: afDEFg).
        """
        # import ipdb;ipdb.set_trace()
        verify_aligned_info(sequence)
        (
            aligned_seq,
            non_active_sites,
            active_sites,
            all_seqs,
        ) = extract_active_sites_info(sequence)

        ans = ''
        for curr_sub_seq in all_seqs:
            for c in curr_sub_seq:
                if (
                    curr_sub_seq[0] <= 'Z'
                ):  # it's uppercase, so it's inside an active site
                    if np.random.rand() < self.mutate_upper:
                        ans += REAL_AAS[np.random.randint(self.num_aas)]
                    else:
                        ans += c
                else:
                    if np.random.rand() < self.mutate_lower:
                        ans += REAL_AAS[np.random.randint(self.num_aas)].lower()
                    else:
                        ans += c

        return ans


class ProteinAugmentSwapSubstrs(Transform):
    """Augment a protein sequence by randomly swapping neighboring subsequences."""

    def __init__(self, p: float = 0.2) -> None:
        """
        Args:
            p (float): Probability that any substr switches places with its "neighbour".

        """
        if not isinstance(p, float):
            raise TypeError(f'Please pass float, not {type(p)}.')
        self._p = np.clip(p, 0.0, 1.0)

    def __call__(self, sequence: str) -> str:
        """
        Apply the transform.

        Args:
            sequence (str): an aligned sequence (example: abCDefGHi).

        Returns:
            str: an aligned sequence with swapped substrings (example: abGHefCDi).
        """
        verify_aligned_info(sequence)
        (
            aligned_seq,
            non_active_sites,
            active_sites,
            all_seqs,
        ) = extract_active_sites_info(sequence)

        order = list(range(len(active_sites)))

        for pos in range(len(order) - 1):
            if np.random.rand() < self._p:
                # switch
                order[pos], order[pos + 1] = order[pos + 1], order[pos]

        curr_active_site_substr_idx = -1
        ans = ''
        for substr in all_seqs:
            if substr[0] <= 'Z':
                curr_active_site_substr_idx += 1
                ans += active_sites[order[curr_active_site_substr_idx]]
            else:
                ans += substr

        return ans



"""Implementation of _SmiEagerDataset."""
from ..files import read_smi
from ..types import Any, Hashable, Sequence
from .dataframe_dataset import DataFrameDataset


class _SmiEagerDataset(DataFrameDataset):
    """
    .smi dataset using eager loading.

    Suggested when handling datasets that can fit in the device memory.
    In case of out of memory errors consider using _SmiLazyDataset.
    """

    def __init__(
        self,
        smi_filepath: str,
        index_col: int = 1,
        name: str = 'SMILES',
        names: Sequence[str] = None,
    ) -> None:
        """
        Initialize a .smi dataset.

        Args:
            smi_filepath (str): path to .smi file.
            index_col (int): Data column used for indexing, defaults to 1.
            name (str): type of dataset, used to index columns in smi, and must
                be in names. Defaults to 'SMILES'.
            names (Sequence[str]): User-assigned names given to the columns.
                Defaults to `[name]`.
        """
        self.smi_filepath = smi_filepath
        self.name = name
        self.names = names or [name]
        self.index_col = index_col
        df = read_smi(self.smi_filepath, index_col=self.index_col, names=self.names)
        DataFrameDataset.__init__(self, df)

    def __getitem__(self, index: int) -> str:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            str: SMILES for the current sample.
        """
        return self.df.iloc[index][self.name]

    def get_item_from_key(self, key: Hashable) -> Any:
        """Get item via key"""
        return self.df.loc[key, self.name]



"""Package installer."""
import codecs
import os

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError('Unable to find version string.')


LONG_DESCRIPTION = ''
if os.path.exists('README.md'):
    with open('README.md') as fp:
        LONG_DESCRIPTION = fp.read()

scripts = ['bin/pytoda-filter-invalid-smi']

setup(
    name='pytoda',
    version=get_version('pytoda/__init__.py'),
    description='pytoda: PaccMann PyTorch Dataset Classes.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Matteo Manica, Jannis Born, Joris Cadow',
    author_email='drugilsberg@gmail.com, jannis.born@gmx.de, joriscadow@gmail.com',
    url='https://github.com/PaccMann/paccmann_datasets',
    license='MIT',
    install_requires=[
        'numpy',
        'scikit-learn',
        'pandas',
        'torch>=1.4.0',
        'diskcache',
        'dill>=0.3.3',
        'selfies>=2.0.0',
        'upfp',
        'SmilesPE>=0.0.3',
        'pyfaidx',
        'pubchempy',
        'importlib_resources',
        'Unidecode',
        'rdkit-pypi>=2021.9.3',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(),
    package_data={
        'pytoda.smiles.metadata': [
            'spe_chembl.txt',
            'ATTRIBUTION',
            'README.md',
            'tokenizer/*',
        ],
        'pytoda.proteins.metadata': ['kinase_activesite_alignment.smi'],
    },
    scripts=scripts,
)



"""Implementation of _TableDataset."""
import copy
from functools import partial

import numpy as np
import torch

from pytoda.warnings import device_warning

from ..types import CsvSourceData, FeatureList, Files, Optional, Tensor
from ._csv_eager_dataset import _CsvEagerDataset
from ._csv_lazy_dataset import _CsvLazyDataset
from ._csv_statistics import reduce_csv_statistics
from .base_dataset import DatasetDelegator
from .utils import concatenate_file_based_datasets


# https://github.com/scikit-learn/scikit-learn/blob/0fb307bf39bbdacd6ed713c00724f8f871d60370/sklearn/preprocessing/_data.py#L63
def _handle_zeros_in_scale(scale, copy=True):
    """
    This method is copied `from sklearn.preprocessing._data`
    Makes sure that whenever scale is zero, we handle it correctly.
    This happens in most scalers when we have constant features.
    """
    # if we are fitting on 1D arrays, scale might be a scalar
    if np.isscalar(scale):
        if scale == 0.0:
            scale = 1.0
        return scale
    elif isinstance(scale, np.ndarray):
        if copy:
            # New array to avoid side-effects
            scale = scale.copy()
        scale[scale == 0.0] = 1.0
        return scale


def transform_not(data: CsvSourceData):
    return data


def transform_standardize(data: CsvSourceData, mean: np.ndarray, std: np.ndarray):
    return (data - mean) / _handle_zeros_in_scale(std, copy=False)


def transform_minmax(data: CsvSourceData, minimum: np.ndarray, maximum: np.ndarray):
    return (data - minimum) / _handle_zeros_in_scale(maximum - minimum, copy=False)


class _TableDataset(DatasetDelegator):
    """
    Table dataset abstract definition.

    The implementation is abstract and can be extend to define different
    data loading policies.
    """

    def __init__(
        self,
        filepaths: Files,
        feature_list: FeatureList = None,
        standardize: bool = True,
        min_max: bool = False,
        processing_parameters: dict = {},
        impute: Optional[float] = None,
        dtype: torch.dtype = torch.float,
        chunk_size: int = 10000,
        device: torch.device = None,
        **kwargs,
    ) -> None:
        """
        Initialize a table dataset.

        Args:
            filepaths (Files): paths to .csv files.
            feature_list (GeneList): a list of features. Defaults to None.
            standardize (bool): perform data standardization. Defaults to True.
            min_max (bool): perform min-max scaling. Defaults to False.
            processing_parameters (dict): processing parameters.
                Keys can be 'min', 'max' or 'mean', 'std'
                respectively. Values must be readable by `np.array`, and the
                required order and subset of features has to match that
                determined by the dataset setup (see `self.feature_list` after
                initialization). Defaults to {}.
            impute (Optional[float]): NaN imputation with value if
                given. Defaults to None.
            dtype (torch.dtype): data type. Defaults to torch.float.
            chunk_size (int): size of the chunks in case of lazy reading, is
                ignored with 'eager' backend. Defaults to 10000.
            device (torch.device): DEPRECATED
            kwargs (dict): additional parameters for pd.read_csv.
        """
        self.filepaths = filepaths
        self.initial_feature_list = feature_list
        self.standardize = standardize
        self.min_max = min_max
        self.processing_parameters = processing_parameters
        self.dtype = dtype
        self.chunk_size = chunk_size
        self.kwargs = copy.deepcopy(kwargs)
        if self.standardize and self.min_max:
            raise RuntimeError('Cannot both standardize and min-max scale')
        self.dataset = None
        self.max = None
        self.min = None
        self.mean = None
        self.std = None
        device_warning(device)

        # the dataset(s) will be initialized individually,
        # the collected statistics will later be updated and finally applied
        # TODO: depending on processing_parameters and standardize/minmax,
        # statistics need not be computed on setup.
        self._setup_dataset()

        # reduce statistics and find definitive feature_list
        (
            self.feature_list,
            self.max,
            self.min,
            self.mean,
            self.std,
        ) = reduce_csv_statistics(self.dataset.datasets, self.initial_feature_list)
        self.number_of_features = len(self.feature_list)

        # given the statistics, we define the transformation
        self.transform_fn = transform_not
        self.processing = {}
        if self.standardize:
            mean = self.processing_parameters.get('mean', self.mean)
            std = self.processing_parameters.get('std', self.std)
            # support scalars, sequences (also of length 1) and arrays
            mean = np.array(mean, dtype=float).squeeze()
            std = np.array(std, dtype=float).squeeze()
            self.transform_fn = partial(transform_standardize, mean=mean, std=std)
            self.processing = {
                'processing': 'standardize',
                'parameters': {'mean': mean.tolist(), 'std': std.tolist()},
            }
        elif self.min_max:
            minimum = self.processing_parameters.get('min', self.min)
            maximum = self.processing_parameters.get('max', self.max)
            # support scalars, sequences (also of length 1) and arrays
            minimum = np.array(minimum, dtype=float).squeeze()
            maximum = np.array(maximum, dtype=float).squeeze()
            self.transform_fn = partial(
                transform_minmax, minimum=minimum, maximum=maximum
            )
            self.processing = {
                'processing': 'min_max',
                'parameters': {'min': minimum.tolist(), 'max': maximum.tolist()},
            }
        # Filter, order and transform the datasets and impute missing values
        for dataset in self.dataset.datasets:
            dataset.transform_dataset(
                transform_fn=self.transform_fn,
                feature_list=self.feature_list,
                impute=impute,
            )

    def _setup_dataset(self) -> None:
        """
        Setup KeyDataset assigned to self.dataset for delegation.
        Defines feature_mapping and fits statistics."""
        raise NotImplementedError

    def __getitem__(self, index: int) -> Tensor:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            torch.tensor: a torch tensor of table values
                for the current sample.
        """
        return torch.tensor(self.dataset[index], dtype=self.dtype)


class _TableLazyDataset(_TableDataset):
    """
    Table dataset using lazy loading.

    Suggested when handling datasets that can fit in the device memory.
    In case of datasets fitting in device memory consider using
    _TableEagerDataset for better performance.
    """

    def _setup_dataset(self) -> None:
        """Setup KeyDataset assigned to self.dataset for delegation."""
        self.dataset = concatenate_file_based_datasets(
            filepaths=self.filepaths,
            dataset_class=_CsvLazyDataset,
            feature_list=self.initial_feature_list,
            chunk_size=self.chunk_size,
            **self.kwargs,
        )


class _TableEagerDataset(_TableDataset):
    """
    Table dataset using eager loading.

    Suggested when handling datasets that can fit in the device memory.
    In case of out of memory errors consider using _TableLazyDataset.
    """

    def _setup_dataset(self) -> None:
        """Setup KeyDataset assigned to self.dataset for delegation."""
        self.dataset = concatenate_file_based_datasets(
            filepaths=self.filepaths,
            dataset_class=_CsvEagerDataset,
            feature_list=self.initial_feature_list,
            **self.kwargs,
        )



"""Testing ProteinProteinInteractionDataset """
import os
import unittest

import numpy as np

from pytoda.datasets import ProteinProteinInteractionDataset
from pytoda.proteins import ProteinFeatureLanguage, ProteinLanguage
from pytoda.tests.utils import TestFileContent


class TestProteinProteinInteractionDataset(unittest.TestCase):
    """Testing annotated dataset."""

    def test___len__(self) -> None:

        content_entity_1 = os.linesep.join(['CCE	ID1', 'KCPR	ID3', 'NCCS	ID2'])
        content_entity_2 = os.linesep.join(
            ['EGK	ID3', 'S	ID1', 'FGAAV	ID2', 'NCCS	ID4']
        )

        annotated_content = os.linesep.join(
            [
                'label_0,label_1,tcr,peptide',
                '2.3,3.4,ID3,ID4',
                '4.5,5.6,ID2,ID1',
                '6.7,7.8,ID1,ID2',
            ]
        )

        with TestFileContent(content_entity_1, suffix='.smi') as a_test_file:
            with TestFileContent(content_entity_2, suffix='.smi') as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:

                    for sequence_filetype in ['.smi', 'infer']:
                        ppi_dataset = ProteinProteinInteractionDataset(
                            [a_test_file.filename, another_test_file.filename],
                            ['tcr', 'peptide'],
                            annotation_file.filename,
                            sequence_filetypes=sequence_filetype,
                        )

                    self.assertEqual(len(ppi_dataset), 3)

        # Test for length if some sequences are not there
        annotated_content = os.linesep.join(
            [
                'label_0,label_1,tcr,peptide',
                '2.3,3.4,ID3,ID4',
                '4.5,5.6,ID2,ID1',
                '6.7,7.8,ID1,ID2',
                '6.7,7.8,ID7,ID2',
                '3.14,1.61,oh,no',
            ]
        )
        with TestFileContent(content_entity_1) as a_test_file:
            with TestFileContent(content_entity_2) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    ppi_dataset = ProteinProteinInteractionDataset(
                        [a_test_file.filename, another_test_file.filename],
                        ['tcr', 'peptide'],
                        annotation_file.filename,
                        sequence_filetypes='.smi',
                    )

                    self.assertEqual(len(ppi_dataset), 3)
                    # ppi_dataset.masks_df to inspect

    def test___getitem__(self) -> None:
        """Test __getitem__."""

        content_entity_1 = os.linesep.join(['CCE	ID1', 'KCPR	ID3', 'NCCS	ID2'])
        content_entity_2 = os.linesep.join(
            ['EGK	ID3', 'S	ID1', 'FGAAV	ID2', 'NCCS	ID4']
        )

        annotated_content = os.linesep.join(
            [
                'label_0,label_1,tcr,peptide',
                '2.3,3.4,ID3,ID4',
                '4.5,5.6,ID2,ID1',
                '6.7,7.8,ID1,ID2',
            ]
        )

        with TestFileContent(content_entity_1) as a_test_file:
            with TestFileContent(content_entity_2) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    ppi_dataset = ProteinProteinInteractionDataset(
                        [a_test_file.filename, another_test_file.filename],
                        ['tcr', 'peptide'],
                        annotation_file.filename,
                        sequence_filetypes='.smi',
                    )

                    # test first sample
                    tcr, peptide, label = ppi_dataset[0]

                    tok_to_idx = ppi_dataset.protein_languages[0].token_to_index

                    self.assertEqual(
                        tcr.numpy().flatten().tolist(),
                        [
                            tok_to_idx['K'],
                            tok_to_idx['C'],
                            tok_to_idx['P'],
                            tok_to_idx['R'],
                        ],
                    )
                    self.assertEqual(
                        peptide.numpy().flatten().tolist(),
                        [
                            tok_to_idx['<PAD>'],
                            tok_to_idx['N'],
                            tok_to_idx['C'],
                            tok_to_idx['C'],
                            tok_to_idx['S'],
                        ],
                    )
                    self.assertTrue(
                        np.allclose(label.numpy().flatten().tolist(), [2.3, 3.4])
                    )

        # Test for non-case-matching entity names
        with TestFileContent(content_entity_1) as a_test_file:
            with TestFileContent(content_entity_2) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    ppi_dataset = ProteinProteinInteractionDataset(
                        [a_test_file.filename, another_test_file.filename],
                        ['TCR', 'Peptide'],
                        annotation_file.filename,
                        sequence_filetypes='.smi',
                    )

                    # test first sample
                    tcr, peptide, label = ppi_dataset[0]

                    tok_to_idx = ppi_dataset.protein_languages[0].token_to_index

                    self.assertEqual(
                        tcr.numpy().flatten().tolist(),
                        [
                            tok_to_idx['K'],
                            tok_to_idx['C'],
                            tok_to_idx['P'],
                            tok_to_idx['R'],
                        ],
                    )
                    self.assertEqual(
                        peptide.numpy().flatten().tolist(),
                        [
                            tok_to_idx['<PAD>'],
                            tok_to_idx['N'],
                            tok_to_idx['C'],
                            tok_to_idx['C'],
                            tok_to_idx['S'],
                        ],
                    )
                    self.assertTrue(
                        np.allclose(label.numpy().flatten().tolist(), [2.3, 3.4])
                    )

        # Switch label columns
        annotated_content = os.linesep.join(
            [
                'label_0,label_1,peptIDE,tcR',
                '2.3,3.4,ID3,ID4',
                '4.5,5.6,ID2,ID1',
                '6.7,7.8,ID1,ID2',
            ]
        )
        content_entity_1 = os.linesep.join(['CCE	ID1', 'KCPR	ID3', 'NCCS	ID2'])
        content_entity_2 = os.linesep.join(
            ['EGK	ID3', 'S	ID1', 'FGAAV	ID2', 'NCCS	ID4']
        )

        with TestFileContent(content_entity_1) as a_test_file:
            with TestFileContent(content_entity_2) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    ppi_dataset = ProteinProteinInteractionDataset(
                        [a_test_file.filename, another_test_file.filename],
                        ['tcr', 'peptide'],
                        annotation_file.filename,
                        sequence_filetypes='.smi',
                    )
                    self.assertEqual(len(ppi_dataset), 2)

                    # test first sample
                    tcr, peptide, label = ppi_dataset[-1]

                    tok_to_idx = ppi_dataset.protein_languages[0].token_to_index
                    self.assertTrue(
                        np.allclose(label.numpy().flatten().tolist(), [6.7, 7.8])
                    )
                    self.assertEqual(
                        tcr.numpy().flatten().tolist(),
                        [
                            tok_to_idx['N'],
                            tok_to_idx['C'],
                            tok_to_idx['C'],
                            tok_to_idx['S'],
                        ],
                    )
                    self.assertEqual(
                        peptide.numpy().flatten().tolist(),
                        [
                            tok_to_idx['<PAD>'],
                            tok_to_idx['<PAD>'],
                            tok_to_idx['<PAD>'],
                            tok_to_idx['<PAD>'],
                            tok_to_idx['S'],
                        ],
                    )

        # Only one annotation column
        with TestFileContent(content_entity_1) as a_test_file:
            with TestFileContent(content_entity_2) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    ppi_dataset = ProteinProteinInteractionDataset(
                        [a_test_file.filename, another_test_file.filename],
                        ['tcr', 'peptide'],
                        annotation_file.filename,
                        sequence_filetypes='.smi',
                        annotations_column_names=['label_0'],
                    )
                    self.assertEqual(len(ppi_dataset), 2)

                    # test first sample
                    tcr, peptide, label = ppi_dataset[-1]

                    tok_to_idx = ppi_dataset.protein_languages[0].token_to_index
                    self.assertTrue(
                        np.allclose(label.numpy().flatten().tolist(), [6.7])
                    )

        # Annotation colum given as index
        with TestFileContent(content_entity_1) as a_test_file:
            with TestFileContent(content_entity_2) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    ppi_dataset = ProteinProteinInteractionDataset(
                        [a_test_file.filename, another_test_file.filename],
                        ['tcr', 'peptide'],
                        annotation_file.filename,
                        sequence_filetypes='.smi',
                        annotations_column_names=[1],
                    )
                    self.assertEqual(len(ppi_dataset), 2)

                    # test first sample
                    tcr, peptide, label = ppi_dataset[-1]

                    tok_to_idx = ppi_dataset.protein_languages[0].token_to_index
                    self.assertTrue(
                        np.allclose(label.numpy().flatten().tolist(), [7.8])
                    )
        # Test for giving only one protein sequence entity
        with TestFileContent(content_entity_2) as a_test_file:
            with TestFileContent(annotated_content) as annotation_file:
                ppi_dataset = ProteinProteinInteractionDataset(
                    [a_test_file.filename],
                    ['peptide'],
                    annotation_file.filename,
                    sequence_filetypes='.smi',
                    annotations_column_names=[1],
                )
                self.assertEqual(len(ppi_dataset), 3)

                # test last sample
                data_tuple = ppi_dataset[-1]
                self.assertEqual(len(data_tuple), 2)

        # Test for giving three entities
        with TestFileContent(content_entity_1) as a_test_file:
            with TestFileContent(content_entity_2) as another_test_file:
                with TestFileContent(content_entity_2) as third_test_file:
                    with TestFileContent(annotated_content) as annotation_file:

                        # Test passing no protein language
                        ppi_dataset = ProteinProteinInteractionDataset(
                            [
                                a_test_file.filename,
                                another_test_file.filename,
                                third_test_file.filename,
                            ],
                            entity_names=['tcr', 'peptide', 'peptide'],
                            labels_filepath=annotation_file.filename,
                            sequence_filetypes='.smi',
                            annotations_column_names=[1],
                        )
                        self.assertEqual(len(ppi_dataset), 2)

                        # test last sample
                        data_tuple = ppi_dataset[-1]
                        self.assertEqual(len(data_tuple), 4)
                        self.assertListEqual(
                            data_tuple[1].tolist(), data_tuple[2].tolist()
                        )

                        # Test passing single protein language
                        for dic in ['iupac', 'unirep', 'human-kinase-alignment']:
                            for s in [False, True]:
                                lang = ProteinLanguage(
                                    amino_acid_dict=dic, add_start_and_stop=s
                                )
                                ppi_dataset = ProteinProteinInteractionDataset(
                                    [
                                        a_test_file.filename,
                                        another_test_file.filename,
                                        third_test_file.filename,
                                    ],
                                    entity_names=['tcr', 'peptide', 'peptide'],
                                    protein_languages=lang,
                                    labels_filepath=annotation_file.filename,
                                    sequence_filetypes='.smi',
                                    annotations_column_names=[1],
                                    add_start_and_stops=s,
                                    iterate_datasets=True,
                                )
                                self.assertEqual(len(ppi_dataset), 2)

                                # test last sample
                                data_tuple = ppi_dataset[-1]
                                self.assertEqual(len(data_tuple), 4)
                                self.assertListEqual(
                                    data_tuple[1].tolist(), data_tuple[2].tolist()
                                )
                                # Tensor should be 1D
                                self.assertEqual(1, len(data_tuple[0].shape))

                        # Test passing single protein_feature_language
                        for dic in ['blosum', 'binary_features', 'float_features']:
                            for s in [False, True]:
                                lang = ProteinFeatureLanguage(
                                    features=dic, add_start_and_stop=s
                                )

                                ppi_dataset = ProteinProteinInteractionDataset(
                                    [
                                        a_test_file.filename,
                                        another_test_file.filename,
                                        third_test_file.filename,
                                    ],
                                    entity_names=['tcr', 'peptide', 'peptide'],
                                    protein_languages=lang,
                                    labels_filepath=annotation_file.filename,
                                    sequence_filetypes='.smi',
                                    annotations_column_names=[1],
                                    add_start_and_stops=s,
                                    iterate_datasets=True,
                                )
                                self.assertEqual(len(ppi_dataset), 2)

                                # test last sample
                                data_tuple = ppi_dataset[-1]
                                self.assertEqual(len(data_tuple), 4)
                                self.assertListEqual(
                                    data_tuple[1].tolist(), data_tuple[2].tolist()
                                )
                                # Tensor should be 2D
                                self.assertEqual(2, len(data_tuple[0].shape))

                        # Test passing mixture of protein and protein_feature languages
                        for dic1 in ['iupac', 'unirep', 'human-kinase-alignment']:
                            for dic2 in ['blosum', 'float_features']:

                                lang1 = ProteinLanguage(amino_acid_dict=dic1)
                                lang2 = ProteinFeatureLanguage(features=dic2)

                                ppi_dataset = ProteinProteinInteractionDataset(
                                    [
                                        a_test_file.filename,
                                        another_test_file.filename,
                                        third_test_file.filename,
                                    ],
                                    entity_names=['tcr', 'peptide', 'peptide'],
                                    protein_languages=[lang1, lang2, lang2],
                                    labels_filepath=annotation_file.filename,
                                    sequence_filetypes='.smi',
                                    annotations_column_names=[1],
                                    add_start_and_stops=True,
                                    iterate_datasets=True,
                                )
                                self.assertEqual(len(ppi_dataset), 2)

                                # test last sample
                                data_tuple = ppi_dataset[-1]
                                self.assertEqual(len(data_tuple), 4)
                                self.assertListEqual(
                                    data_tuple[1].tolist(), data_tuple[2].tolist()
                                )
                                # Tensor dimensions
                                for i, g in zip(range(len(data_tuple) - 1), (1, 2, 2)):
                                    self.assertEqual(g, len(data_tuple[i].shape))

                                # Testing alternative ordering
                                ppi_dataset = ProteinProteinInteractionDataset(
                                    [
                                        a_test_file.filename,
                                        another_test_file.filename,
                                        third_test_file.filename,
                                    ],
                                    entity_names=['tcr', 'peptide', 'peptide'],
                                    protein_languages=[lang2, lang1, lang2],
                                    labels_filepath=annotation_file.filename,
                                    sequence_filetypes='.smi',
                                    annotations_column_names=[1],
                                    add_start_and_stops=True,
                                    iterate_datasets=True,
                                )
                                self.assertEqual(len(ppi_dataset), 2)

                                # test last sample
                                data_tuple = ppi_dataset[-2]
                                self.assertEqual(len(data_tuple), 4)
                                # Test whether decoding with PL and PFL gives the same
                                feat1 = data_tuple[1].long().tolist()
                                if dic2 == 'blosum':
                                    feat0 = [
                                        tuple(x) for x in data_tuple[0].long().tolist()
                                    ]

                                    feat2 = [
                                        tuple(x) for x in data_tuple[2].long().tolist()
                                    ]
                                elif dic2 == 'float_features':
                                    feat0 = [
                                        tuple([round(xx, 2) for xx in x])
                                        for x in data_tuple[0].double().tolist()
                                    ]
                                    feat2 = [
                                        tuple([round(xx, 2) for xx in x])
                                        for x in data_tuple[2].double().tolist()
                                    ]

                                self.assertEqual(
                                    'CCE',
                                    ppi_dataset.protein_languages[
                                        0
                                    ].token_indexes_to_sequence(feat0),
                                )
                                self.assertEqual(
                                    'FGAAV',
                                    ppi_dataset.protein_languages[
                                        1
                                    ].token_indexes_to_sequence(feat1),
                                )
                                self.assertEqual(
                                    'FGAAV',
                                    ppi_dataset.protein_languages[
                                        2
                                    ].token_indexes_to_sequence(feat2),
                                )
                                # Tensor dimensions
                                for i, g in zip(range(len(data_tuple) - 1), (2, 1, 2)):
                                    self.assertEqual(g, len(data_tuple[i].shape))

        # Test for using different padding lengths
        padding_lengths = [8, 6]
        with TestFileContent(content_entity_1) as a_test_file:
            with TestFileContent(content_entity_2) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    ppi_dataset = ProteinProteinInteractionDataset(
                        [
                            a_test_file.filename,
                            another_test_file.filename,
                        ],
                        ['tcr', 'peptide'],
                        annotation_file.filename,
                        padding_lengths=padding_lengths,
                        sequence_filetypes='.smi',
                    )
                    self.assertEqual(len(ppi_dataset), 2)

                    # test last sample
                    data_tuple = ppi_dataset[-1]
                    self.assertEqual(len(data_tuple), 3)
                    for i, p in enumerate(padding_lengths):
                        self.assertEqual(len(data_tuple[i]), p)


if __name__ == '__main__':
    unittest.main()



"""Testing .smi preprocessing utilities."""
import os
import unittest
from io import StringIO

from pytoda.preprocessing.smi import filter_invalid_smi, find_undesired_smiles_files
from pytoda.tests.utils import TestFileContent


class TestSmi(unittest.TestCase):
    """Testing .smi preprocessing."""

    def test_filter_invalid_smi(self) -> None:
        """Test filter_invalid_smi."""
        smiles_content = os.linesep.join(['CCO	compound_a', 'C(	compound_b'])
        filtered_smiles_content = os.linesep.join(['CCO	compound_a'])
        with TestFileContent(smiles_content) as smiles_file:
            with TestFileContent(filtered_smiles_content) as filtered_smiles_file:
                with TestFileContent('') as resulting_smiles_file:
                    print(
                        "\nExpected 'SMILES Parse Error' while filtering "
                        "invalid smiles via rdkit:"
                    )
                    filter_invalid_smi(
                        smiles_file.filename, resulting_smiles_file.filename
                    )
                    with open(resulting_smiles_file.filename) as result_fp:
                        with open(filtered_smiles_file.filename) as filtered_fp:
                            self.assertEqual(
                                result_fp.read().strip(),
                                filtered_fp.read().strip(),
                            )

    def test_find_undesired_smiles_files(self) -> None:
        """Test find_undesired_smiles_files."""

        UNDESIRED = os.linesep.join(['CCO	CHEMBL545', 'NCCS	CHEMBL602'])
        MORE_UNDESIRED = os.linesep.join(['NC(=O)O	CHEMBL123', 'NCCS	CHEMBL602'])
        CONTENT = os.linesep.join(
            [
                'SMILES,ID',
                'COCC(C)N,CHEMBL3184692',
                'COCCOC,CHEMBL1232411',
                'O=CC1CCC1,CHEMBL18475',
                'NC(=O)O,CHEMBL125278',
            ]
        )

        for undesired, gt in zip(
            [UNDESIRED, MORE_UNDESIRED],
            [
                'No matches found, shutting down.\n',
                'Found NC(=O)O in list of undesired SMILES.\n',
            ],
        ):
            with TestFileContent(CONTENT) as content:
                with TestFileContent(undesired) as undesired_content:

                    mystdout = StringIO()
                    find_undesired_smiles_files(
                        undesired_content.filename,
                        content.filename,
                        save_matches=False,
                        file=mystdout,
                    )
                    self.assertEqual(mystdout.getvalue(), gt)


if __name__ == '__main__':
    unittest.main()



"""Initialize the dataset module."""
from .annotated_dataset import AnnotatedDataset  # noqa
from .base_dataset import ConcatKeyDataset, DatasetDelegator, KeyDataset  # noqa
from .distributional_dataset import DistributionalDataset  # noqa
from .drug_affinity_dataset import DrugAffinityDataset  # noqa
from .drug_sensitivity_dataset import DrugSensitivityDataset  # noqa
from .drug_sensitivity_dose_dataset import DrugSensitivityDoseDataset  # noqa
from .gene_expression_dataset import GeneExpressionDataset  # noqa
from .polymer_dataset import PolymerTokenizerDataset  # noqa
from .protein_protein_interaction_dataset import (  # noqa
    ProteinProteinInteractionDataset,
)
from .protein_sequence_dataset import ProteinSequenceDataset  # noqa
from .set_matching_dataset import (  # noqa
    PairedSetMatchingDataset,
    PermutedSetMatchingDataset,
    SetMatchingDataset,
)
from .smiles_dataset import SMILESDataset, SMILESTokenizerDataset  # noqa
from .utils import indexed, keyed  # noqa



"""Implementation of ProteinProteinInteractionDataset."""
import pandas as pd
import torch
from numpy import iterable
from torch.utils.data import Dataset

from pytoda.warnings import device_warning

from ..proteins.protein_language import ProteinLanguage
from ..types import Files, List, Sequence, Tensor, Tuple, Union
from .protein_sequence_dataset import ProteinSequenceDataset


class ProteinProteinInteractionDataset(Dataset):
    """
    PPI Dataset implementation. Designed for two sources of protein sequences
    and on source of discrete labels.
    NOTE: Only supports classification (possibly multitask) but no regression
    tasks.
    """

    def __init__(
        self,
        sequence_filepaths: Union[Files, Sequence[Files]],
        entity_names: Sequence[str],
        labels_filepath: str,
        sequence_filetypes: Union[str, List[str]] = 'infer',
        annotations_column_names: Union[List[int], List[str]] = None,
        protein_languages: Union[ProteinLanguage, List[ProteinLanguage]] = None,
        paddings: Union[bool, Sequence[bool]] = True,
        padding_lengths: Union[int, Sequence[int]] = None,
        add_start_and_stops: Union[bool, Sequence[bool]] = False,
        augment_by_reverts: Union[bool, Sequence[bool]] = False,
        randomizes: Union[bool, Sequence[bool]] = False,
        iterate_datasets: Union[bool, Sequence[bool]] = False,
        device: torch.device = None,
    ) -> None:
        """
        Initialize a protein protein interactiondataset.

        Args:
            sequence_filepaths (Union[Files, Sequence[Files]]):
                paths to .smi (also as .csv) or .fasta (.gz) file for protein
                sequences. For each item in the iterable, one protein sequence
                dataset is created. Sequences can be nested, i.e. each protein
                sequence dataset can be created from an iterable of filepaths
                of same type, see sequence_filetypes.
            entity_names (Sequence[str]): List of protein sequence entities,
                e.g. ['Peptides', 'T-Cell-Receptors']. These names should be
                column names of the labels_filepaths in order respective to
                sequence_filepaths.
            labels_filepath (str): path to .csv file with classification
                labels.
            sequence_filetypes (Union[str, List[str]]): the filetypes of the
                sequence files. Can either be a str if all files have identical
                types or an Sequence if different entities have different
                types. Different types across the same entity are not
                supported. Supported formats are {.smi, .csv, .fasta,
                .fasta.gz}. Default is `infer`, i.e. filetypes are inferred
                automatically.
            annotations_column_names (Union[List[int], List[str]]): indexes
                (positional or strings) for the annotations. Defaults to None,
                a.k.a. all the columns, except the entity_names are annotation
                labels.
            protein_languages (Union[ProteinLanguage, List[ProteinLanguage]):
                one or multiple ProteinLanguages. If multiple are provided, exactly one
                should be given for each entity. If only one is provided, the same
                language will be used for all entities. You can also pass child instances
                like ProteinFeatureLanguage. Defaults to None, i.e., creating a single
                protein language with iupac dictionary for all entities.
            paddings (Union[bool, Sequence[bool]]): pad sequences to longest in
                the protein language. Defaults to True.
            padding_lengths (Union[int, Sequence[int]]): manually sets number
                of applied paddings (only if padding = True). Defaults to None.
            add_start_and_stops (Union[bool, Sequence[bool]]): add start and
                stop token indexes.  Defaults to False.
            augment_by_reverts (Union[bool, Sequence[bool]]): perform a
                stochastic reversion of the amino acid sequence.
            randomizes (Union[bool, Sequence[bool]]): perform a true
                randomization of the amino acid sequences. Defaults to False.
            iterate_datasets (Union[bool, Sequence[bool]]): whether to go through all
                items in the datasets to detect unknown characters, find longest
                sequence and checks passed padding length if applicable.
                Defaults to False.
            device (torch.device): DEPRECATED

        """
        Dataset.__init__(self)
        device_warning(device)
        assert len(entity_names) == len(
            sequence_filepaths
        ), 'sequence_filepaths should be an iterable of length in entity names'

        self.labels_filepath = labels_filepath
        self.entities = list(map(lambda x: x.capitalize(), entity_names))

        # wrap single filepath per entity to treat equally as iterable (*args)
        self.sequence_filepaths = [
            [filepath] if isinstance(filepath, str) else filepath
            for filepath in sequence_filepaths
        ]
        #  Data type of first sequence files per entity
        if sequence_filetypes == 'infer':
            self.filetypes = list(
                map(lambda x: '.' + x[0].split('.')[-1], self.sequence_filepaths)
            )
        elif sequence_filetypes in ['.smi', '.csv', '.fasta', '.fasta.gz']:
            self.filetypes = [sequence_filetypes] * len(self.entities)
        elif len(sequence_filetypes) == len(self.entities) and all(
            x in ['.smi', '.csv', '.fasta', '.fasta.gz'] for x in sequence_filetypes
        ):
            self.filetypes = sequence_filetypes
        else:
            raise ValueError(f'Unsupported filetype: {sequence_filetypes}')

        (
            self.paddings,
            self.padding_lengths,
            self.add_start_and_stops,
            self.augment_by_reverts,
            self.randomizes,
            self.iterate_datasets,
        ) = map(
            (
                lambda x: x
                if iterable(x) and len(x) == len(self.entities)
                else [x] * len(self.entities)
            ),
            (
                paddings,
                padding_lengths,
                add_start_and_stops,
                augment_by_reverts,
                randomizes,
                iterate_datasets,
            ),
        )

        if protein_languages is None:
            # Objects will be constructed in `ProteinSequenceDataset`
            self.protein_languages = [None for _ in self.entities]
            self.pl_methods = ['iupac' for _ in self.entities]
        else:
            if isinstance(protein_languages, Sequence):
                # Multiple languages were passed
                self.protein_languages = protein_languages
                self.pl_methods = [p.method for p in protein_languages]

            elif isinstance(protein_languages, ProteinLanguage):
                # Single language was passed
                self.protein_languages = [protein_languages for _ in self.entities]
                self.pl_methods = [protein_languages.method for _ in self.entities]

            else:
                raise TypeError(
                    f'Received unknown type for protein_language: {type(protein_languages)}'
                )
            # Check whether input arguments are consistent
            for i, (lang, start) in enumerate(
                zip(self.protein_languages, self.add_start_and_stops)
            ):
                assert (
                    lang.add_start_and_stop == start
                ), f'add_start_and_stop differs for language {i}: {start} vs. {lang.add_start_and_stop}'

        # Create protein sequence datasets.
        self.datasets = [
            ProteinSequenceDataset(
                *filepaths,
                filetype=self.filetypes[index],
                protein_language=self.protein_languages[index],
                padding=self.paddings[index],
                padding_length=self.padding_lengths[index],
                add_start_and_stop=self.add_start_and_stops[index],
                augment_by_revert=self.augment_by_reverts[index],
                randomize=self.randomizes[index],
                name=self.entities[index],
                iterate_dataset=self.iterate_datasets[index],
            )
            for index, filepaths in enumerate(self.sequence_filepaths)
        ]
        # Retrieve the possibly updated protein languages
        self.protein_languages = [data.protein_language for data in self.datasets]

        # Labels
        self.labels_df = pd.read_csv(self.labels_filepath)
        # Cast the column names to uppercase
        self.labels_df.columns = map(
            lambda x: str(x).capitalize(), self.labels_df.columns
        )
        columns = self.labels_df.columns

        # handle labels
        if annotations_column_names is None:
            self.labels = [column for column in columns if column not in self.entities]
        elif all([isinstance(column, int) for column in annotations_column_names]):
            self.labels = columns[annotations_column_names]
        elif all([isinstance(column, str) for column in annotations_column_names]):
            self.labels = list(map(lambda x: x.capitalize(), annotations_column_names))
        else:
            raise RuntimeError(
                'label_columns should be an iterable containing int or str'
            )
        # get the number of labels
        self.number_of_tasks = len(self.labels)

        assert all(
            list(map(lambda x: x in self.labels_df.columns, self.entities))
        ), 'At least one given entity name was not found in labels_filepath.'

        # filter data based on the availability
        masks = []
        mask = pd.Series([True] * len(self.labels_df), index=self.labels_df.index)

        for entity, dataset in zip(self.entities, self.datasets):
            # prune rows (in mask) with ids unavailable in respective dataset
            local_mask = self.labels_df[entity].isin(set(dataset.keys()))
            mask = mask & local_mask
            masks.append(local_mask)

        self.labels_df = self.labels_df.loc[mask]

        # to investigate missing ids per entity
        self.masks_df = pd.concat(masks, axis=1)
        self.masks_df.columns = self.entities

        self.number_of_samples = len(self.labels_df)

    def __len__(self) -> int:
        "Total number of samples."
        return self.number_of_samples

    def __getitem__(self, index: int) -> Tuple[Tensor, ...]:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            Tuple: a tuple containing self.entities+1 torch.Tensors
            representing respectively: compound token indexes for each
            protein entity and the property labels (annotations)
        """

        # sample selection
        selected_sample = self.labels_df.iloc[index]

        # labels (annotations)
        labels_tensor = torch.tensor(
            list(selected_sample[self.labels].values),
            dtype=torch.float,
        )
        # samples (Protein sequences)
        proteins_tensors = [
            ds.get_item_from_key(selected_sample[ds.name]) for ds in self.datasets
        ]
        return tuple([*proteins_tensors, labels_tensor])
